#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

import operator
from .dict_util import reverse_dict

def __in__(x: object, y: iter) -> bool:
    """
    Checks whether ``x`` is in ``y``.

    Args:
        x (object): An item.
        y (iter): A container.

    Returns:
        ``True`` iff ``x in y``,
        ``False`` otherwise.
    """
    return operator.__contains__(y, x)

OPERATORS = {
    ">"   : operator.__gt__,
    "<"   : operator.__lt__,
    ">="  : operator.__ge__,
    "<="  : operator.__le__,
    "=="  : operator.__eq__,
    "!="  : operator.__ne__,
    "||"  : operator.__or__,
    "OR"  : operator.__or__,
    "&&"  : operator.__and__,
    "AND" : operator.__and__,
    "^"   : operator.__xor__,
    "XOR" : operator.__xor__,
    "IN"  : __in__,
    "CONTAINS" : operator.__contains__,
}

OPERATORS_TO_STR = reverse_dict(OPERATORS)

class BinaryPredicate:
    """
    The :py:class:`BinaryPredicate` represents a binary predicate
    to define minifold queries.
    See also the :py:class:`Query` class (especially, the ``filters`` attribute).
    """
    def __init__(self, left, operator, right):
        """
        Constructor.

        Example:
            >>> from minifold import BinaryPredicate
            >>> bp = BinaryPredicate("a", "<=", 0)
            >>> entry = {"a": 1, "b": 2}
            >>> bp(entry)
            False
            >>> entry = {"a": -1, "b": 2}
            >>> bp(entry)
            True

        Args:
            left (str): The left operand. Maybe the key of an entry.
            operator (str): The binary operator, see also :py:data:`OPERATORS`.
            right (str): The right operand. Cannot be the key of an entry.
        """
        self.m_left     = left
        self.m_operator = OPERATORS.get(operator, operator)
        self.m_right    = right
        assert not isinstance(self.m_operator, str)

    @property
    def left(self) -> str:
        """
        Retrieves the left operand of this :py:class:`BinaryPredicate`.

        Returns:
            The left operand of this :py:class:`BinaryPredicate`.
        """
        return self.m_left

    @property
    def operator(self) -> str:
        """
        Retrieves the binary operator of this :py:class:`BinaryPredicate`.

        Returns:
            The binary operator of this :py:class:`BinaryPredicate`.
        """
        return self.m_operator

    @property
    def right(self):
        """
        Retrieves the right operand of this :py:class:`BinaryPredicate`.

        Returns:
            The right operand of this :py:class:`BinaryPredicate`.
        """
        return self.m_right

    def __str__(self) -> str:
        """
        Converts this :py:class:`BinaryPredicate` to a string.

        Returns:
            The string representing this :py:class:`BinaryPredicate`.
        """
        return "%s %s %s" % (self.left, OPERATORS_TO_STR[self.operator], self.right)

    def match(self, entry: dict) -> bool:
        """
        Matches an entry against this :py:class:`BinaryPredicate`.

        Args:
            entry (dict): An entry. One of its key must corresponds to ``self.left``.

        Raises:
            KeyError: if the entry does not have key matching ``self.left``.

        Returns:
            ``True`` if this :py:class:`BinaryPredicate` is satisfied by ``entry``,
            ``False`` otherwise.
        """
        try:
            if self.operator in [operator.__or__, operator.__and__, operator.__xor__]:
                return self.operator(self.left(entry), self.right(entry))
            else:
                left = entry[self.left]
                right = self.right
                return self.operator(left, right)
        except KeyError:
            return (self.operator == operator.__eq__ and self.right == None)

    def __call__(self, entry: dict) -> bool:
        """
        Short-hand for the :py:meth:`BinaryPredicate.match` method.

        Args:
            entry (dict): An entry. One of its key must corresponds to ``self.left``.

        Raises:
            KeyError: if the entry does not have key matching ``self.left``.

        Returns:
            ``True`` if this :py:class:`BinaryPredicate` is satisfied by ``entry``,
            ``False`` otherwise.
        """
        return self.match(entry)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

from copy import deepcopy
from .binary_predicate import BinaryPredicate
from .dict_util import reverse_dict
from .connector import Connector
from .query import Query

def rename_list(l: list, mapping: dict) -> list:
    """
    Replaces some values involved in a list.

    Example:
        >>> rename_list(["a", "b", "a", "c"], {"a": "A", "b": "B"})
        ['A', 'B', 'A', 'c']

    Args:
        l (list): The input list, modified in place.
        mapping (dict): A dictionary mapping each value to
            be replaced by the new corresponding value.

    Returns:
        The updated list.
    """
    for (i, value) in enumerate(l):
        try:
            l[i] = mapping[value]
        except KeyError as e:
            pass  # see test_dblp.py
    return l

def rename_key(d: dict, old_key: str, new_key: str) -> dict:
    """
    Replaces a key (possibly) involved in a dictionary.

    Example:
        >>> rename_key({"a": 1, "b": 2, "c": 3}, "a", "A")
        {'b': 2, 'c': 3, 'A': 1}

    Args:
        d (dict): The input dictionary, updated in place.
        old_key (str): The key to be updated.
        new_key (str): The new key.

    Returns:
        The updated dictionary.
    """
    try:
        d[new_key] = d.pop(old_key)
    except KeyError:
        pass
    return d

def rename_entry(d: dict, mapping: dict) -> dict:
    """
    Replaces several keys (possibly) involved in a dictionary.

    Example:
        >>> rename_entry({"a": 1, "b": 2, "c": 3}, {"a": "A", "b": "B"})
        {'c': 3, 'A': 1, 'B': 2}

    Args:
        d (dict): The input dictionary, updated in place.
        mapping (dict): A dictionary mapping each key to
            be replaced by the new corresponding key.

    Returns:
        The updated dictionary.
    """
    for (old_key, new_key) in mapping.items():
        rename_key(d, old_key, new_key)
    return d

def rename(mapping: dict, entries: list) -> list:
    """
    Replaces several keys (possibly) involved in a list of minifold entries.

    Example:
        >>> rename({"a": "A", "b": "B"}, [{"a": 1, "b": 2, "c": 3}, {"a": 10, "b": 20, "c": 30}])
        [{'c': 3, 'A': 1, 'B': 2}, {'c': 30, 'A': 10, 'B': 20}]

    Args:
        mapping (dict): A dictionary mapping each key to
            be replaced by the new corresponding key.
        entries (list): A list of minifold entries, updated in place.

    Returns:
        The updated entries.
    """
    for entry in entries:
        rename_entry(entry, mapping)
    return entries

# TODO minifold.searchSearchFilter is not supported
def rename_filters(filters: object, mapping: dict):
    """
    Rename the keys involved in a list of minifold filters
    (see :py:class:`BinaryPredicate`), typically the WHERE
    part of a :py:class:`Query` instance.
    Recursive function.
    See the :py:func:`rename_query` function.

    Args:
        filters (object): The minifold filters, updated in place.
        mapping (dict): A dictionary mapping each key to
            be replaced by the new corresponding key.


    Returns:
        The updated filters.
    """
    if filters is None or isinstance(filters, int) or isinstance(filters, tuple):
        pass
    elif isinstance(filters, str):
        try:
            filters = mapping[filters]
        except KeyError:
            pass
    elif isinstance(filters, BinaryPredicate):
        filters.m_left  = rename_filters(filters.left,  mapping)
        filters.m_right = rename_filters(filters.right, mapping)
    else:
        raise RuntimeError("rename_filters: unsupported type %s: filters = %s" % (type(filters), filters))
    return filters

def rename_sort_by(sort_by: dict, mapping: dict) -> dict:
    """
    Rename the SORT BY part of a :py:class:`Query` instance.
    See the :py:func:`rename_query` function.

    Args:
        sort_by (dict): The SORT BY part of a :py:class:`Query` instance.
        mapping (dict): A dictionary mapping each key to
            be replaced by the new corresponding key.

    Returns:
        The dictionary corresponding to ``sort_by`` after renaming.
    """
    return {
        mapping.get(attribute, attribute): sort_asc
        for attribute, sort_asc in sort_by.items()
    } if sort_by else dict()

def rename_query(q: Query, mapping: dict) -> Query:
    """
    Rename some attributes involved in a :py:class:`Query` instance.
    This is especially useful when a minifold pipeline involves several
    data sources using different naming conventions for a same attribute.

    Args:
        q (Query); A py:class:`Query` instance.
        mapping (dict): A dictionary mapping each key to
            be replaced by the new corresponding key.

    Returns:
        The renamed minifold query.
    """
    q_renamed = q
    rename_list(q_renamed.attributes, mapping)
    rename_filters(q_renamed.filters, mapping)
    q.sort_by = rename_sort_by(q_renamed.sort_by, mapping)
    return q_renamed

class RenameConnector(Connector):
    """
    The :py:class:`RenameConnector` class wraps the :py:func:`rename` function
    to exploit it in a minifold query plan.
    """
    def __init__(self, mapping: dict = None, child: Connector = None):
        """
        Constructor.

        Args:
            mapping (dict): A dictionary mapping each key to
                be replaced by the new corresponding key
                (overlying to underlying ontology).
                Note that mapping must be reversible, i.e.
                ``reverse_dict(reverse_dict(d)) == d``.
            child (Connector): The child minifold :py:class:`Connector`
                instance.
        """
        super().__init__()
        self.m_map_rq = mapping if mapping else dict()
        self.m_map_qr = reverse_dict(mapping)
        self.m_child = child

    def attributes(self, object: str) -> set:
        """
        Lists the available attributes related to a given collection of
        minifold entries exposed by this :py:class:`RenameConnector` instance.

        Args:
            object (str): The name of the collection.

        Returns:
            The set of corresponding attributes.
        """
        return {
            self.map_rq.get(attribute, attribute)
            for attribute in self.m_child.attributes(object)
        }

    def query(self, q: Query) -> list:
        """
        Handles an input :py:class:`Query` instance.

        Args:
            query (Query): The handled query.

        Returns:
            The list of entries matching the input query.
        """
        super().query(q)
        assert self.child != None
        q_renamed = rename_query(deepcopy(q), self.map_qr)
        return self.answer(
            q,
            rename(
                self.map_rq,
                self.child.query(q_renamed)
            )
        )

    @property
    def child(self):
        """
        Accessor to the child minifold :py:class:`Connector` instance.

        Returns:
            The child minifold :py:class:`Connector` instance.
        """
        return self.m_child

    @property
    def map_qr(self):
        """
        Accessor to the renaming mapping (overlying to underlying).

        Returns:
            The corresponding mapping.
        """
        return self.m_map_qr

    @property
    def map_rq(self):
        """
        Accessor to the renaming mapping (underlying to overlying).

        Returns:
            The corresponding mapping.
        """
        return self.m_map_rq



#!/usr/bin/env python3
#
# Twitter connector
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

import datetime
from .connector import Connector
from .query import ACTION_READ, Query

try:
    import tweepy
except ImportError as e:
    from .log import Log
    Log.warning(
        "Please install requests-cache.\n"
        "  APT: sudo apt install python3-tweepy\n"
        "  PIP: sudo pip3 install --upgrade tweepy\n"
    )
    raise e

def tweet_to_dict(tweet) -> dict:
    """
    Converts a tweet to a dictionary.

    Args:
        tweet: The input tweet.

    Returns:
        The corresponding dictionary.
    """
    # Fetch tweet
    try:
        content = tweet._json["full_text"]
    except:
        content = tweet._json["text"]
    if content:
        content = content.split(" https://t.co/")[0]

    # Get tweet image
    image = None
    try:
        image = tweet._json["entities"]["media"][0]["media_url"]
    except:
        # Try to get image from youtube, e.g: https://img.youtube.com/vi/iZCyv0e2RLM/1.jpg
        try:
            if "youtu" in tweet._json["entities"]["urls"][0]["display_url"]:
                image = (
                    "https://img.youtube.com/vi/%s/1.jpg"
                    % tweet._json["entities"]["urls"][0]["display_url"].split("/")[-1]
                )
        except:
            Log.warning("No image found for tweet %s" % content)

    # Get author name
    author_name = None
    try:
        author_name = tweet._json["user"]["name"]
    except:
        Log.warning("No author name found for tweet %s" % content)

    # Get author image
    author_image = None
    try:
        author_image = tweet._json["user"]["profile_image_url"]
    except:
        Log.warning("No author image found for tweet %s" % content)

    # Get tweet date
    date_list = tweet._json["created_at"].split()
    display_date = "%(m)s %(d)s, %(y)s" % {
        "d" : date_list[2],
        "m" : date_list[1],
        "y" : date_list[5]
    }
    date = datetime.datetime.strptime(display_date, "%b %d, %Y")
    display_date = date.strftime("%d/%m/%Y")

    # Return entry
    return {
        "text"         : content,
        "image"        : image,
        "date"         : date,
        "display_date" : display_date,
        "author_name"  : author_name,
        "author_image" : author_image,
    }

class TwitterConnector(Connector):
    """
    The :py:class:`TwitterConnector` is a gateway minifold allowing
    to fetch tweets from Twitter.
    """
    def __init__(
        self,
        twitter_id: str,
        consumer_key: str,
        consumer_secret: str,
        access_token: str,
        access_token_secret: str
    ):
        """
        Constructor.

        Args:
            twitter_id (str): The Twitter identifier.
            consumer_key (str): The Twitter API Key.
            consumer_secret (str): The Twitter secret.
            access_token (str): The Twitter access token.
            access_token_secret: The Twitter access token secret.
        """
        super().__init__()
        self.twitter_id          = twitter_id
        self.consumer_key        = consumer_key
        self.consumer_secret     = consumer_secret
        self.access_token        = access_token
        self.access_token_secret = access_token_secret
        self.api                 = None
        self.connect()

    def attributes(self, object: str) -> set:
        """
        Lists the available attributes related to a given collection of
        minifold entries exposed by this :py:class:`WhereConnector` instance.

        Args:
            object (str): The name of the collection.
                Valid collections are ``"self"`` and ``"feed"``.

        Returns:
            The set of corresponding attributes.
        """

        return {
            "text",
            "image",
            "date",
            "display_date",
            "author_name",
            "author_image"
        } if object in {"self", "feed"} else set()

    def connect(self):
        """
        Connect to the Twitter API.
        """
        auth = tweepy.OAuthHandler(self.consumer_key, self.consumer_secret)
        auth.set_access_token(self.access_token, self.access_token_secret)
        self.api = tweepy.API(auth)

    def query(self, q: Query) -> list:
        """
        Handles an input :py:class:`Query` instance.

        Args:
            query (Query): The handled query.

        Returns:
            The list of entries matching the input query.
        """
        if not self.api:
            raise RuntimeError("Not connected to twitter")

        if q.action == ACTION_READ:
            if q.object in {"self", "feed"}:
                # Retrieve Tweets
                assert q.limit > 0
                size = q.limit
                if q.object == "self":
                    tweets = self.api.user_timeline(
                        id = self.twitter_id,
                        tweet_mode = "extended",
                        count = size
                    )
                elif q.object == "feed":
                    tweets = self.api.home_timeline(
                        id = self.twitter_id,
                        tweet_mode = "extended",
                        count = size
                    )
                else:
                    tweets = list()
                entries = [tweet_to_dict(tweet) for tweet in tweets]
            else:
                raise ValueError(f"Invalid object '{q.object}'")
        else:
            raise RuntimeError(f"Query {q}: action {q.action} not implemented")
        return self.answer(q, entries)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# If you need more generic join, see join_if.py
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

from .connector import Connector
from .query     import Query
from .join_if   import merge_dict

def are_naturally_joined(l: dict, r: dict) -> bool:
    """
    Internal function, used to check whether two dictionaries can be joined
    using an NATURAL JOIN statement.

    Args:
        l (dict): The dictionary corresponding to the left operand.
        r (dict): The dictionary corresponding to the right operand.

    Returns:
        ``True`` if and only if ``l`` and ``r`` can be joined,
        ``False`` otherwise (i.e., ``l`` and ``r`` have no common key).
    """
    inter_keys = set(l.keys()) & set(r.keys())
    if len(inter_keys) == 0:
        return False
    for key in inter_keys:
        if l[key] != r[key]:
            return False
    return True

def natural_join(l_entries: list, r_entries: list) -> list:
    """
    Computes the NATURAL JOIN of two lists of minifold entries.

    Args:
        l_entries (dict): The minifold entries corresponding to the left operand.
        r_entries (dict): The minifold entries corresponding to the right operand.
        f (callable): A functor such that ``f(l, r)`` returns
            ``True`` if and only if ``l`` and ``r`` can be joined
            (where ``l`` and ``r`` are two minifold entries),
            ``False`` otherwise.
        match_once (bool): Pass ``True`` if a left entry must be matched
            at most once.
        merge (callable): A function that merges two input dictionaries.
            Defaults to :py:func:`merge_dict`.

    Returns:
        The corresponding list of entries.
    """
    ret = list()
    for l in l_entries:
        for r in r_entries:
            if are_naturally_joined(l, r):
                ret.append(merge_dict(l, r))
    return ret

class NaturalJoinConnector(Connector):
    """
    The :py:class:`NaturalJoinConnector` is a minifold connector that implements
    the NATURAL JOIN statement in a minifold pipeline.
    """
    def __init__(self, left: Connector, right: Connector):
        """
        Constructor.

        Args:
            left (Connector): The left :py:class:`Connector` child.
            right (Connector): The right :py:class:`Connector` child.
        """
        super().__init__()
        self.m_left = left
        self.m_right = right
        self.m_left_entries = list()
        self.m_right_entries = list()

    def attributes(self, object: str) -> set:
        """
        Lists the available attributes related to a given collection of
        minifold entries exposed by this :py:class:`JoinIfConnector` instance.

        Args:
            object (str): The name of the collection.

        Returns:
            The set of corresponding attributes.
        """
        return (
            self.m_left.attributes(object) |
            self.m_right.attributes(object)
        )

    def query(self, query: Query) -> list:
        """
        Handles an input :py:class:`Query` instance.

        Args:
            query (Query): The handled query.

        Returns:
            The list of entries matching the input query.
        """

        self.m_left_entries.clear()
        self.m_right_entries.clear()
        self.m_left_entries = self.left.query(query)
        self.m_right_entries = self.right.query(query)
        return self.answer(
            query,
            natural_join(self.m_left_entries, self.m_right_entries)
        )

    @property
    def left(self) -> Connector:
        """
        Retrieves the left :py:class:`Connector` child
        in this :py:class:`JoinIfConnector` instance.

        Returns:
            The left :py:class:`Connector` child.
        """
        return self.m_left

    @property
    def right(self) -> Connector:
        """
        Retrieves the right :py:class:`Connector` child
        in this :py:class:`JoinIfConnector` instance.

        Returns:
            The left :py:class:`Connector` child.
        """
        return self.m_right



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

from .connector import Connector
from .query import Query
from .values_from_dict import ValuesFromDictFonctor

def sort_by_impl(functor: ValuesFromDictFonctor, entries: list, desc: bool = True) -> list:
    """
    Implementation details of :func:`sort_by`.

    Args:
        functor (ValuesFromDictFonctor): The functor allowing to extract
            the values used to sort the entries.
        entries (list): A list of minifold entries.
        desc (bool): Pass ``True`` to sort by ascending order,
            ``False`` otherwise.

    Returns:
        The sorted entries, with respect to ``functor``.
    """
    return sorted(entries, key=functor, reverse=desc)

def sort_by(attributes :list, entries :list, desc: bool = False) -> list:
    """
    Sorts a list of minifold entries.

    Args:
        attributes (list): The list of entry keys used to sort.
        entries (list): A list of minifold entries.
        desc (bool): Pass ``True`` to sort by ascending order,
            ``False`` otherwise.

    Returns:
        The sorted entries, with respect to ``functor``.
    """
    functor = ValuesFromDictFonctor(attributes)
    return sort_by_impl(functor, entries, desc)

class SortByConnector(Connector):
    """
    The :py:class:`SortByConnector` class implements the SORT BY
    statement in a minifold pipeline.
    """
    def __init__(self, attributes: list, child: Connector, desc: bool = False):
        """
        Constructor.

        Args:
            attributes (list): The list of entry keys used to sort.
            child (Connector): The child minifold :py:class:`Connector`
                instance.
        desc (bool): Pass ``True`` to sort by ascending order,
            ``False`` otherwise.
        """
        super().__init__()
        self.m_functor = ValuesFromDictFonctor(attributes)
        self.m_child = child
        self.m_desc = desc

    def attributes(self, object :str) -> set:
        """
        Lists the available attributes related to a given collection of
        minifold entries exposed by this :py:class:`SortByConnector` instance.

        Args:
            object (str): The name of the collection.

        Returns:
            The set of corresponding attributes.
        """
        return self.child.attributes(object)

    @property
    def child(self):
        """
        Accessor to the child minifold :py:class:`Connector` instance.

        Returns:
            The child minifold :py:class:`Connector` instance.
        """
        return self.m_child

    @property
    def desc(self) -> bool:
        """
        Checks whether this :py:class:`SortByConnector` sorts the entries
        by descending order.

        Returns:
            ``True`` if this :py:class:`SortByConnector` sorts the entries
            by descending order, ``False`` otherwise.
        """
        return self.m_desc

    def query(self, q :Query) -> list:
        """
        Handles an input :py:class:`Query` instance.

        Args:
            query (Query): The handled query.

        Returns:
            The list of entries matching the input query.
        """
        super().query(q)
        return self.answer(
            q,
            sort_by_impl(
                self.m_functor,
                self.m_child.query(q),
                self.m_desc
            )
        )

    def __str__(self) -> str:
        """
        Returns the string representation of this
        :py:class:`SortByConnector` instance

        Returns:
            The string representation of this
            :py:class:`SortByConnector` instance
        """
        return "SORT BY %s %s" % (
            ", ".join(self.m_functor.attributes),
            "DESC" if self.desc else "ASC"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

import io, csv
from enum import IntEnum

from .connector import Connector
from .query import Query, ACTION_READ, action_to_str
from .log import Log

class CsvModeEnum(IntEnum):
    """
    Enumeration listing the different kind of supported CSV data source.
    """
    FILENAME = 0  # Use text file
    STRING   = 1  # Use str
    TEXTIO   = 2  # Use any class inheriting TextIOBase

    def __str__(self) -> str:
        s = IntEnum.__str__(self)
        i = len(self.__class__.__name__)
        return s[i + 1:]

class CsvConnector(Connector):
    """
    The :py:class:`CsvConnector` is a minifold gateway allowing to manipulate
    data stored in CSV file.
    """
    def __init__(
        self,
        data: str,
        delimiter: chr = ' ',
        quotechar: chr = '"',
        mode: CsvModeEnum = CsvModeEnum.FILENAME
    ):
        """
        Constructor.

        Args:
            data (str): Depending on the nature of the CSV data source,
                this string either contains the path to the CSV file;
                or the CSV data itself; or the ``TextIOBase`` instance.
                See the ``mode`` parameter.
            delimiter (str): The string that delimits each column of the
                input CSV. Example: ``";"``, ``"|"``, ``" "``.
                Defaults to ``' '``.
            quotechar (str): The charater used to delimits values
                that may contain ``delimiter``. Defaults to ``'"'``.
            mode (CsvModeEnum): The nature of the input CSV source.
                See also the :py:class:`CsvModeEnum` enumeration.
        """
        super().__init__()
        stream = (
            open(data, "rt")  if mode == CsvModeEnum.FILENAME else
            io.StringIO(data) if mode == CsvModeEnum.STRING   else
            data              if mode == CsvModeEnum.TEXTIO   else
            None
        )
        if not stream:
            raise RuntimeError("CsvConnector: Invalid input stream (data = %r, mode = %s)" % (data, mode))

        # csv.reader cannot directly process a str, so we use a io.StringIO
        # See https://stackoverflow.com/questions/3305926/python-csv-string-to-array
        Log.debug("delimiter = %r, quotechar = %r" % (delimiter, quotechar))
        Log.debug(type(stream))
        reader = csv.reader(stream, delimiter=delimiter, quotechar=quotechar)
        rows = [row for row in reader]
        Log.debug(rows)

        if mode == CsvModeEnum.FILENAME and stream:
            stream.close()

        # Assuming that attributes are declared in the first line of the CSV data.
        self.indexed_attributes = rows[0]
        self.entries = [
            {self.indexed_attributes[i] : v
            for (i, v) in enumerate(row)} for row in rows[1:]
        ]

    def attributes(self, object: str):
        """
        Lists the attributes of the collection of objects stored
        in this :py:class:`CsvConnector` instance.

        Args:
            object (str): The name of the minifold object.
                As a :py:class:`CsvConnector` instance stores a single
                collection, ``object`` is no relevant and you may pass ``None``.

        Returns:
            The set of available ``object``'s attributes
        """
        return set(self.indexed_attributes)

    def query(self, query: Query) -> list:
        """
        Handles an input :py:class:`Query` instance.

        Args:
            query (Query): The handled query.

        Returns:
            The list of entries matching ``query``.
        """
        super().query(query)
        if query.action == ACTION_READ:
            ret = self.reshape_entries(query, self.entries)
        else:
            raise RuntimeError(
                "CsvConnector.query: %s not yet implemented" % action_to_str(query.action)
            )
        return self.answer(query, ret)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project
# https://github.com/nokia/minifold

"""Top-level package."""

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"
__version__ = '0.10.0'  # Use single quotes for bumpversion (see setup.cfg)

from .binary_predicate import OPERATORS, OPERATORS_TO_STR, BinaryPredicate
from .cached import CachedEntriesConnector, JsonCachedConnector, PickleCachedConnector
from .cache import CacheConnector, JsonCacheConnector, PickleCacheConnector
from .closure import is_multiple_key, closure, minimal_cover
from .config import DEFAULT_MINIFOLD_CONFIG, Config
from .connector import Connector
from .connector_util import show_some_values
from .count import CountConnector, count, count_gen
from .country import country_code_to_name
from .csv import CsvConnector, CsvModeEnum
from .dblp import DBLP_API_URL, DBLP_ALIASES, DblpConnector
from .dict_util import reverse_dict, freeze_dict
from .doc_type import DocType, doc_type_to_html
from .download import DownloadConnector, download, downloads, now, trim_http
from .entries_connector import EntriesConnector
from .filesystem import check_writable_directory, ctime, find, mkdir, mtime, rm
from .for_each import ForEachFilter, for_each_sub_entry
from .google_scholar import GoogleScholarConnector
from .group_by import GroupByConnector, group_by
from .hal import HAL_API_URL, HAL_ALIASES, HalConnector
from .html import (
    connector_to_html, entry_to_html, entries_to_html, html, print_error, value_to_html
)
from .html_table import HtmlTableConnector, HtmlTableParser, html_table
from .ipynb import in_ipynb
from .join_if import (
    INNER_JOIN, LEFT_JOIN, RIGHT_JOIN, FULL_OUTER_JOIN, merge_dict,
    JoinIfConnector, inner_join_if, left_join_if, right_join_if, full_outer_join_if
)
from .json import JsonConnector, JsonFileConnector
from .lambdas import LambdasConnector, lambdas
from .ldap import LdapConnector
from .lexical_cast import cast_bool, cast_none, lexical_cast, lexical_casts
from .limit import LimitConnector, limit
from .log import (
    RED, GREEN, YELLOW, BLUE, PINK, CYAN, GRAY,
    BOLD, UNDERLINED, BLINKING, HIGHLIGHTED,
    Log
)
from .mongo import MongoConnector
from .natural_join import NaturalJoinConnector, are_naturally_joined, natural_join
from .proxy import Proxy, proxy_enable, proxy_disable, make_session, proxy_enable_localhost
from .query import (
    ACTION_CREATE, ACTION_READ, ACTION_UPDATE, ACTION_DELETE,
    SORT_ASC, SORT_DESC, Query
)
from .rename import RenameConnector, rename, rename_query
from .request_cache import install_cache
from .search import (
    SearchFilter, search,
    equals, contains, lower_case_contains, lower_case_equals, contains_words
)
from .select import SelectConnector, select
from .singleton import Singleton
from .sort_by import SortByConnector, sort_by
from .strings import (
    to_international_string, remove_punctuation,
    remove_html_tags, remove_html_escape_sequences,
    remove_latex_escape_sequence,
    to_canonic_string, to_canonic_fullname, unicode_to_utf8
)
from .twitter import TwitterConnector, tweet_to_dict
from .union import UnionConnector, union
from .unique import UniqueConnector, unique
from .unnest import UnnestConnector, unnest
from .where import WhereConnector, where



#!/usr/bin/env python3
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

from html.parser import HTMLParser
from .log import Log
from .entries_connector import EntriesConnector

class HtmlTableParser(HTMLParser):
    """
    The :py:class:`HtmlTableParser` class extracts the values from an HTML table.
    """
    def __init__(self, columns: list, output_list: list, keep_entry: callable = None):
        """
        Constructor.

        Args:
            columns (list): A list of string mapping the attribute name
                corresponding with the index. If data is fetch
                for columns having a greater index than
                len(columns), columns[-1] is used, and this
                key may store a list of string values instead
                of a single string. This allow to store data
                stored among several columns in a single attribute.
            output_list (list): A reference to an output list where the
                data will be output (one dict per row, one
                key/value per column).
            keep_entry (callable): A callback which determines whether an
                must entry must be kept or discarded. Pass ``None``
                to filter nothing. This is the opportunity to
                discard a header or irrelevant rows.
        """
        HTMLParser.__init__(self)
        self.fetch_data = False
        self.columns    = columns
        self.index      = 0
        self.entries    = output_list
        self.entry      = dict()
        self.value      = str()
        self.keep_entry = keep_entry

    # Inherited abstract method
    def error(self, message: str):
        """
        Logs an error message.

        Args:
            message (str): The error message.
        """
        Log.error(message)

    def handle_starttag(self, tag: str, attrs: str):
        """
        Callback that handles an opening HTML tag (e.g., ``"<td>"``).

        Args:
            tag (str): The HTML tag. _Example:_ ``"td"``.
            attrs (str): The HTML tag attributes.
        """
        if tag == "td":
            # Enable fetch data
            self.fetch_data = True

    def handle_endtag(self, tag):
        """
        Callback that handles an opening HTML tag (e.g., ``"</td>"``).

        Args:
            tag (str): The HTML tag. _Example:_ ``"td"``.
        """
        if tag == "td":  # Push key/value
            # Disable fetch data
            self.fetch_data = False

            # Push new key/value pair
            key = (
                self.columns[self.index] if self.index < len(self.columns)
                else self.columns[-1]
            )
            if key in self.entry.keys():
                current_value = self.entry[key]
                if not isinstance(current_value, list):
                    self.entry[key] = [current_value]
                if self.value:
                    self.entry[key].append(self.value)
            else:
                if self.value:
                    self.entry[key] = self.value

            # Reset key/value pair
            self.value = str()
            self.index += 1
        elif tag == "tr":
            # Push entry
            if self.keep_entry is None or self.keep_entry(self.entry):
                self.entries.append(self.entry)

            # Reset entry
            self.index = 0
            self.entry = dict()

    def handle_data(self, data: str):
        """
        Callback that handles an opening HTML data.

        Args:
            data (str): The HTML data (here, stored in a table cell).
        """
        data = data.strip()
        if self.fetch_data and data:
            self.value += data

def html_table(filename: str, columns: list, keep_entry: bool = None) -> list:
    """
    Loads an HTML table from an input file

    Args:
        filename (str): The path to the input HTML file.
        columns (list): A list of string mapping the attribute name
            corresponding with the index. If data is fetch
            for columns having a greater index than
            len(columns), columns[-1] is used, and this
            key may store a list of string values instead
            of a single string. This allow to store data
            stored among several columns in a single attribute.
        keep_entry (callable): A callback which determines whether an
            must entry must be kept or discarded. Pass ``None``
            to filter nothing. This is the opportunity to
            discard a header or irrelevant rows.

    Returns:
        The corresponding list of minifold entries.
    """
    entries = list()
    parser = HtmlTableParser(columns, entries, keep_entry)
    with open(filename, "r") as f:
        s = f.read()
        parser.feed(s)
    return entries

class HtmlTableConnector(EntriesConnector):
    """
    The :py:class:`HtmlTableConnector` class is a minifold gateway allowing
    to fetch data stored in an HTML table.
    """
    def __init__(self, filename :str, columns :list, keep_entry = None):
        """
        Constructor.

        Args:
            filename: Input HTML filename.
            columns: list of string mapping the attribute name
                corresponding with the index. If data is fetch
                for columns having a greater index than
                len(columns), columns[-1] is used, and this
                key may store a list of string values instead
                of a single string. This allow to store data
                stored among several columns in a single attribute.
            keep_entry: callback which determine whether an
                must entry must be kept or discard. Pass None
                to filter nothing. This is the opportunity to
                discard a header or irrelevant row.
        """
        super().__init__(list())
        self.m_parser = HtmlTableParser(columns, self.m_entries, keep_entry)
        with open(filename, "r") as f:
            s = f.read()
            self.m_parser.feed(s)
        super().__init__(self.m_entries)

    def attributes(self, object: str = None) -> set:
        """
        Lists available attributes related to a given collection of
        minifold entries exposed by this :py:class:`Connector` instance.

        Args:
            object (str): The name of the collection.
                As this connector stores a single collection, you may pass ``None``.
                Defaults to ``None``.

        Returns:
            The set of corresponding attributes.
        """
        return set(self.m_parser.columns)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

from .connector import Connector
from .hash import to_hashable
from .query import Query
from .values_from_dict import ValuesFromDictFonctor

def unique_impl(functor :ValuesFromDictFonctor, entries :list) -> list:
    """
    Implementation details of :func:`unique`.

    Args:
        functor (ValuesFromDictFonctor): The functor allowing to extract
            the values used to form the aggregates.
        entries (list): A list of minifold entries.

    Returns:
        A dictionary where each entries are unique with respect to ``functor``.
    """
    ret = list()
    seen_keys = set()
    for entry in entries:
        key = functor(entry)
        if len(key) == 1:
            (key,) = key
        key = to_hashable(key)
        if key not in seen_keys:
            ret.append(entry)
            seen_keys.add(key)
    return ret

def unique(attributes :list, entries :list) -> list:
    """
    Implements the UNIQUE statement for a list of minifold entries.

    Args:
        attributes (list): The list of entry keys used to determine the uniqueness.
        entries (list): A list of minifold entries.

    Returns:
        The remaining entries once the UNIQUE filtering has been applied.
    """
    functor = ValuesFromDictFonctor(attributes)
    return unique_impl(functor, entries)

class UniqueConnector(Connector):
    """
    The :py:class:`UniqueConnector` class implements the GROUP BY
    statement in a minifold pipeline.
    """
    def __init__(self, attributes :list, child):
        """
        Constructor.

        Args:
            attributes (list): The list of entry keys used to form
                the aggregates.
            child (Connector): The child minifold :py:class:`Connector`
                instance.
        """
        super().__init__()
        self.m_functor = ValuesFromDictFonctor(attributes)
        self.m_child = child

    @property
    def child(self):
        """
        Accessor to the child minifold :py:class:`Connector` instance.

        Returns:
            The child minifold :py:class:`Connector` instance.
        """
        return self.m_child

    def attributes(self, object :str) -> set:
        """
        Lists the available attributes related to a given collection of
        minifold entries exposed by this :py:class:`UniqueConnector` instance.

        Args:
            object (str): The name of the collection.

        Returns:
            The set of corresponding attributes.
        """

        return self.child.attributes(object)

    def query(self, q :Query) -> list:
        """
        Handles an input :py:class:`Query` instance.

        Args:
            query (Query): The handled query.

        Returns:
            The list of entries matching the input query.
        """
        super().query(q)
        return self.answer(
            q,
            unique_impl(
                self.m_functor,
                self.m_child.query(q)
            )
        )

    def __str__(self) -> str:
        """
        Returns the string representation of this
        :py:class:`UniqueConnector` instance

        Returns:
            The string representation of this
            :py:class:`UniqueConnector` instance
        """
        return "DUP %s" % ", ".join(self.m_functor.attributes)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

import sys
from pprint import pformat
from .query import Query
from .log import Log

class Connector:
    """
    The :py:class:`Connector` class is the base class of most of classes involved in
    minifold.

    A minifold query plan is a hierarchy of :py:class:`Connector` instances.
    This hierarchy forms a pipeline, where leaves corresponds to gateways to
    some data sources and internal nodes to SQL-like operators or intermediate
    caches.

    Running a query consists in sending a :py:class:`Query` instance to the root
    :py:class:`Connector` instance of the query plan.
    Then, each :py:class:`Connector` (possibly alters) and forwards
    the query to its children.
    The leaves of the query plan are :py:class:`Connector` are gateways allowing
    to fetch data from a remote or local data source.
    A read query returns the entries matched by the input minifold query (if any)
    to the parent node. Iteratively, the entries reach the root node of the query plan.

    As a result, once the hierarchy of the query plan is ready, the only relevant
    entry point is the root :py:class:`Connector` instance.
    """
    trace_queries = False
    trace_entries = False
    trace_only_keys = False

    def __init__(self):
        """
        Constructor.
        """
        pass

    if sys.version_info >= (3, 6):
        subclasses = dict()

        # https://stackoverflow.com/questions/5189232/how-to-auto-register-a-class-when-its-defined
        def __init_subclass__(cls, **kwargs):
            super().__init_subclass__(**kwargs)
            name = repr(cls).split("'")[1]
            Connector.subclasses[name] = cls

        @staticmethod
        def get_class(name: str):
            cls = Connector.subclasses.get(name)
            if not cls:
                raise RuntimeError(
                    "Invalid connector type [%s]. Known types are:\n%s\n" % (
                        name,
                        "\n\t".join([str(name) for name in Connector.subclasses.keys()])
                    )
                )
            return cls

    def query(self, query: Query) -> list:
        """
        Handles an input :py:class:`Query` instance.

        Args:
            query (Query): The handled query.

        Returns:
            The list of entries matching the input query.
        """
        if Connector.trace_queries:
            Log.debug("%r: --> %s" % (self, query))
        # This method must be overloaded in child classes as follows:
        # super().query(query)
        # entries = [ ... ]
        # return self.answer(query, entries)
        return list()

    def attributes(self, object: str) -> set:
        """
        Lists the available attributes related to a given collection of
        minifold entries exposed by this :py:class:`Connector` instance.

        Args:
            object (str): The name of the collection.

        Returns:
            The set of corresponding attributes.
        """
        # Must be overwritten in child class if reshape_entries is needed
        raise NotImplementedError

    def reshape_entries(self, query: Query, entries: list) -> list:
        """
        Reshapes entries returned by :py:meth:`self.query` before calling
        :py:meth:`self.answer`.

        This method should only be called if the Connector only support a subset
        of query operators among ``{SELECT, WHERE, LIMIT, OFFSET}`` in SQL.

        Args:
            query (Query): The handled :py:class:`Query` instance.
            entries (list): The list of raw entries fetched so far, corresponding to
                ``SELECT * FROM foo LIMIT n WHERE n >= query.limit``.

        Returns:
            The reshaped entries.
        """
        max_attributes = self.attributes(query.object)
        attributes = (
            set(query.attributes) & max_attributes if query.attributes
            else max_attributes
        )

        ret = list()
        for entry in entries[query.offset:]:
            # LIMIT
            if len(ret) == query.limit:
                break

            # WHERE
            if query.filters is None or query.filters(entry):
                # SELECT
                entry = {k : v for (k, v) in entry.items() if k in attributes}
                missing_attributes = set(attributes) - set(entry.keys())
                for k in missing_attributes:
                    entry[k] = None

                ret.append(entry)
        return ret

    def answer(self, query: Query, ret: list):
        """
        Method traversed when this :py:class:`Connector` is ready to
        answer to a given :py:class:`Query`.

        This method is used in the child classes to trace entries in
        the (complex) query plans.

        Args:
            query (Query): The related :py:class:`Query` instance.
            ret (list): The corresponding results.
        """
        if Connector.trace_entries or Connector.trace_only_keys:
            if isinstance(ret, list) and Connector.trace_only_keys and len(ret) > 0:
                entry = ret[0]
                message = "Forwarding %d entries, keys = {%s}" % (
                    len(ret),
                    ", ".join(entry.keys())
                )
            else:
                message = pformat(ret)
            Log.debug("%r: <-- %s\n%s" % (self, query, message))
        return ret



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

from .connector import Connector
from .query import Query

def select(entries :list, attributes :list) -> list:
    """
    Implements the SELECT statement for a list of minifold entries.

    Example:
        >>> select([{"a": 1, "b": 2, "c": 3}, {"a": 10, "b": 20, "c": 30}], ["a", "b"])
        [{'a': 1, 'b': 2}, {'a': 10, 'b': 20}]

    Args:
        entries (list): A list of minifold entries.
        attributes (list): The selected keys.

    Returns:
        The input entries, restricted to the key of interest.
    """
    return [{k : entry[k] for k in attributes} for entry in entries]

class SelectConnector(Connector):
    """
    The :py:class:`SelectConnector` class implements the SELECT
    statement in a minifold pipeline.
    """
    def __init__(self, child: Connector, attributes: list):
        """
        Constructor.

        Args:
            child (Connector): The child minifold :py:class:`Connector`
                instance.
            attributes (list): The selected keys.
        """
        super().__init__()
        self.m_child = child
        self.m_attributes = attributes

    def attributes(self, object :str) -> set:
        """
        Lists the available attributes related to a given collection of
        minifold entries exposed by this :py:class:`SelectConnector` instance.

        Args:
            object (str): The name of the collection.

        Returns:
            The set of corresponding attributes.
        """
        return set(self.m_attributes) & self.m_child.attributes(object)

    @property
    def child(self):
        """
        Accessor to the child minifold :py:class:`Connector` instance.

        Returns:
            The child minifold :py:class:`Connector` instance.
        """
        return self.m_child

    def query(self, query :Query) -> list:
        """
        Handles an input :py:class:`Query` instance.

        Args:
            query (Query): The handled query.

        Returns:
            The list of entries matching the input query.
        """
        super().query(query)
        q = query.copy()
        q.attributes = [
            attribute
            for attribute in q.attributes
            if attribute in self.m_attributes
        ]
        return self.answer(
            query,
            select(
                self.m_child.query(q),
                self.m_attributes
            )
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

from .connector import Connector
from .query import Query

def where(entries: list, f: callable) -> list:
    """
    Implements the WHERE statement for a list of minifold entries.

    Example:
        >>> where([{"a": 1, "b": 1}, {"a": 2, "b": 2}, {"a": 3, "b": 3}], lambda e: e["a"] <= 2)
        [{'a': 1, 'b': 1}, {'a': 2, 'b': 2}]

    Args:
        entries (list): A list of minifold entries.
        f (callable): A function such that ``f(entry)`` returns
            ``True`` if ``entry`` must be kept,
            ``False`` otherwise.

    Returns:
        The kept entries.
    """
    return [entry for entry in entries if f(entry)]

class WhereConnector(Connector):
    """
    The :py:class:`WhereConnector` class implements the WHERE
    statement in a minifold pipeline.
    """
    def __init__(self, child: Connector, keep_if: callable):
        """
        Constructor.

        Args:
            child (Connector): The child minifold :py:class:`Connector`
                instance.
            keep_if (callable): A function such that ``f(entry)`` returns
                ``True`` if ``entry`` must be kept,
                ``False`` otherwise.
        """
        super().__init__()
        self.m_child = child
        self.m_keep_if = keep_if

    @property
    def child(self):
        """
        Accessor to the child minifold :py:class:`Connector` instance.

        Returns:
            The child minifold :py:class:`Connector` instance.
        """
        return self.m_child

    def attributes(self, object :str) -> set:
        """
        Lists the available attributes related to a given collection of
        minifold entries exposed by this :py:class:`WhereConnector` instance.

        Args:
            object (str): The name of the collection.

        Returns:
            The set of corresponding attributes.
        """
        return self.child.attributes(object)

    @property
    def keep_if(self):
        """
        Accessor to the filtering function used by this
        :py:class:`WhereConnector` instance.

        Returns:
            The child minifold :py:class:`Connector` instance.
        """
        return self.m_keep_if

    def query(self, q :Query) -> list:
        """
        Handles an input :py:class:`Query` instance.

        Args:
            query (Query): The handled query.

        Returns:
            The list of entries matching the input query.
        """
        super().query(q)
        return self.answer(
            q,
            where(
                self.m_child.query(q),
                self.m_keep_if
            )
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

def cast_none(s: str) -> None:
    """
    Casts a string to ``None`` if possible, raises an exception otherwise.

    Args:
        s (str): The string to be casted.

    Raises:
        ValueError: if the cast cannot be achieved.

    Returns:
        ``None`` if successful.
    """
    if s is None:
        return None
    elif isinstance(s, str) and s.lower() == "none":
        return None
    raise ValueError("Invalid literal for cast_none(): %s" % s)

def cast_bool(s: str) -> bool:
    """
    Casts a string to a ``bool`` if possible, raises an exception otherwise.

    Args:
        s (str): The string to be casted.

    Raises:
        ``ValueError``, if the cast cannot be achieved.

    Returns:
        The boolean corresponding to s if successful.
    """
    if isinstance(s, bool):
        return s
    elif isinstance(s, str):
        l = s.lower()
        if l == "true":
            return True
        elif l == "false":
            return False
    raise ValueError("Invalid literal for cast_bool(): %s" % s)

def lexical_cast(s: str, cast: callable) -> object:
    """
    Casts a string according to an operator.
    See also the :py:func:`lexical_casts` function.

    Args:
        s (str): The string to be casted.
        cast (callable): A single cast operator.
            _Examples:_ :py:func:`cast_bool`, :py:func:`cast_none`,
            :py:func:`int`, :py:func:`float`, etc.

    Raises:
        ``ValueError``, if the cast cannot be achieved.

    Returns:
        The corresponding value.
    """
    return cast(s)

def lexical_casts(s: str, cast_operators: list = None) -> object:
    """
    Casts a string according to several cast operators.
    See also the :py:func:`lexical_cast` function.

    Args:
        s (str): The string to be casted.
        cast_operators: A list of cast operators. Operators must be ordered
            by decreasing strictness (e.g. ``int`` should preceed ``float``).
            Pass ``None`` to use the default list of cast operators.

    Returns:
        The original string if no cast worked, the corresponding casted value
        otherwise.
    """
    if cast_operators is None:
        cast_operators = [cast_none, cast_bool, int, float]
    for cast in cast_operators:
        try:
            ret = lexical_cast(s, cast)
            return ret
        except ValueError:
            pass
    return s



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

from .connector import Connector
from .query import Query

"""
This file gathers the functions and the classes related the joins supported
in minifold (excepted the natural join) corresponding in SQL to
the INNER JOIN, LEFT JOIN, RIGHT JOIN
and FULL OUTER JOIN statements.

`This schema <http://4.bp.blogspot.com/-_HsHikmChBI/VmQGJjLKgyI/AAAAAAAAEPw/JaLnV0bsbEo/s1600/sql%2Bjoins%2Bguide%2Band%2Bsyntax.jpg>` recalls their meaning.

See also:
- the :py:class:`NaturalJoinConnector` class;
- the :py:func:`natural_join` function.
"""

INNER_JOIN = 1  # Returns matching records
LEFT_JOIN = 2  # Returns left entries + matching right entries (=> right attributes may be None)
RIGHT_JOIN = 3  # Returns right entries + matching left entries (=> left attributes may be None)
FULL_OUTER_JOIN = 4  # Returns left and right entries while merging matching entries

def merge_dict(l :dict, r :dict) -> dict:
    """
    Merges two dictionaries. The input dictionaries are not altered.

    Example:
        >>> merge_dict({"a": 1, "b": 2}, {"b": 3, "c": 4})
        {'a': 1, 'b': 3, 'c': 4}

    Args:
        l (dict): A dictionary.
        r (dict): Another dictionary.

    Returns:
        The dictionary obtained by merging ``l`` and ``r``.
    """
    ret = l.copy()
    ret.update(r)
    return ret

def join_mode_to_string(mode: int) -> str:
    """
    Convert a ``*_JOIN`` constant to the corresponding string representation.

    Args:
    """
    if mode == INNER_JOIN:
        return "INNER JOIN"
    elif mode == LEFT_JOIN:
        return "LEFT JOIN"
    elif mode == RIGHT_JOIN:
        return "RIGHT JOIN"
    elif mode == FULL_OUTER_JOIN:
        return "FULL OUTER JOIN"
    return "?"

def are_joined_if(l: dict, r: dict, f: callable) -> bool:
    """
    Internal function, used to check whether two dictionaries can be joined
    according to a functor.

    Args:
        l (dict): The dictionary corresponding to the left operand.
        r (dict): The dictionary corresponding to the right operand.
        f (callable): A functor such that ``f(l, r)`` returns
            ``True`` if and only if ``l`` and ``r`` can be joined,
            ``False`` otherwise.

    Raises:
        :py:class:`RuntimeError`, if ``f`` applies
            on keys that are missing in ``l`` or ``r``.

    Returns:
        ``True`` if and only if ``l`` and ``r`` can be joined,
        ``False`` otherwise.
    """
    try:
        return f(l, r)
    except KeyError as e:
        raise RuntimeError("Cannot join\n\tl = %r\n\tr = %r\n%s" % (l, r, e))

def inner_join_if(
    l_entries: list,
    r_entries: list,
    f: callable,
    match_once: bool = True,
    merge: callable = merge_dict
) -> list:
    """
    Computes the INNER JOIN of two lists of minifold entries.

    Args:
        l_entries (dict): The minifold entries corresponding to the left operand.
        r_entries (dict): The minifold entries corresponding to the right operand.
        f (callable): A functor such that ``f(l, r)`` returns
            ``True`` if and only if ``l`` and ``r`` can be joined
            (where ``l`` and ``r`` are two minifold entries),
            ``False`` otherwise.
        match_once (bool): Pass ``True`` if a left entry must be matched
            at most once.
        merge (callable): A function that merges two input dictionaries.
            Defaults to :py:func:`merge_dict`.

    Returns:
        The corresponding list of entries.
    """
    ret = list()
    for l in l_entries:
        for r in r_entries:
            if are_joined_if(l, r, f):
                ret.append(merge(l, r))
                if match_once:
                    break
    return ret

def left_join_if(
    l_entries :list,
    r_entries :list, f,
    match_once = True,
    merge = merge_dict
) -> list:
    """
    Computes the LEFT JOIN of two lists of minifold entries.

    Args:
        l_entries (dict): The minifold entries corresponding to the left operand.
        r_entries (dict): The minifold entries corresponding to the right operand.
        f (callable): A functor such that ``f(l, r)`` returns
            ``True`` if and only if ``l`` and ``r`` can be joined
            (where ``l`` and ``r`` are two minifold entries),
            ``False`` otherwise.
        match_once (bool): Pass ``True`` if a left entry must be matched
            at most once.
        merge (callable): A function that merges two input dictionaries.
            Defaults to :py:func:`merge_dict`.

    Returns:
        The corresponding list of entries.
    """
    ret = list()
    if len(r_entries) > 0:
        for l in l_entries:
            joined = False
            r = None
            for r in r_entries:
                if are_joined_if(l, r, f):
                    joined = True
                    ret.append(merge(l, r))
                    if match_once:
                        break
            if joined is False:
                entry = l
                missing_keys = set(r.keys()) - set(l.keys())
                for k in missing_keys:
                    entry[k] = None
                ret.append(entry)
    else:
        ret = l_entries
    return ret

def right_join_if(l_entries :list, r_entries :list, f, match_once = True) -> list:
    """
    Computes the RIGHT JOIN of two lists of minifold entries.

    Args:
        l_entries (dict): The minifold entries corresponding to the left operand.
        r_entries (dict): The minifold entries corresponding to the right operand.
        f (callable): A functor such that ``f(l, r)`` returns
            ``True`` if and only if ``l`` and ``r`` can be joined
            (where ``l`` and ``r`` are two minifold entries),
            ``False`` otherwise.
        match_once (bool): Pass ``True`` if a left entry must be matched
            at most once.

    Returns:
        The corresponding list of entries.
    """
    return left_join_if(r_entries, l_entries, lambda l,r: f(r,l), match_once)

def full_outer_join_if(l_entries :list, r_entries :list, f, match_once = True) -> list:
    """
    Computes the FULL OUTER JOIN of two lists of minifold entries.

    Args:
        l_entries (dict): The minifold entries corresponding to the left operand.
        r_entries (dict): The minifold entries corresponding to the right operand.
        f (callable): A functor such that ``f(l, r)`` returns
            ``True`` if and only if ``l`` and ``r`` can be joined
            (where ``l`` and ``r`` are two minifold entries),
            ``False`` otherwise.
        match_once (bool): Pass ``True`` if a left entry must be matched
            at most once.

    Returns:
        The corresponding list of entries.
    """
    # Get every left entries (join-able or not)
    ret = left_join_if(l_entries, r_entries, f, match_once)

    # Retrieve left keys
    l_keys = set()
    for l in l_entries:
        l_keys |= {k for k in l.keys()}

    # Get missing right entries
    for r in r_entries:
        joined = False
        for l in l_entries:
            if are_joined_if(l, r, f):
                joined = True
                break

        if not joined:
            entry = r
            missing_keys = set(l_keys) - set(r.keys())
            for k in missing_keys:
                entry[k] = None
            ret.append(entry)

    return ret

class JoinIfConnector(Connector):
    """
    The :py:class:`JoinIfConnector` is a minifold connector that implements
    the INNER JOIN, LEFT JOIN, RIGHT JOIN
    and FULL OUTER JOIN statements in a minifold pipeline.
    """
    def __init__(
        self,
        left: Connector,
        right: Connector,
        join_if: callable,
        mode: int = INNER_JOIN
    ):
        """
        Constructor.

        Args:
            left (Connector): The left :py:class:`Connector` child.
            right (Connector): The right :py:class:`Connector` child.
            join_if (callable): The callback implementing the join criterion.
            mode (int): The type of join. The valid values are:
                :py:data:`INNER_JOIN`,
                :py:data:`LEFT_JOIN`,
                :py:data:`RIGHT_JOIN`,
                :py:data:`FULL_OUTER_JOIN`.
        """
        super().__init__()
        self.m_left = left
        self.m_right = right
        self.m_join_if = join_if
        self.m_left_entries = list()
        self.m_right_entries = list()
        self.m_mode = mode

    def attributes(self, object :str):
        """
        Lists the available attributes related to a given collection of
        minifold entries exposed by this :py:class:`JoinIfConnector` instance.

        Args:
            object (str): The name of the collection.

        Returns:
            The set of corresponding attributes.
        """
        return (
            self.m_left.attributes(object) |
            self.m_right.attributes(object)
        )

    def query(self, query: Query) -> list:
        """
        Handles an input :py:class:`Query` instance.

        Args:
            query (Query): The handled query.

        Raises:
            :py:class:`ValueError` if :py:attr:`self.m_mode` is not valid.

        Returns:
            The list of entries matching the input query.
        """
        super().query(query)
        self.m_left_entries.clear()
        self.m_right_entries.clear()
        self.m_left_entries = self.left.query(query)
        self.m_right_entries = self.right.query(query)
        if self.m_mode == INNER_JOIN:
            entries = inner_join_if(
                self.m_left_entries,
                self.m_right_entries,
                self.m_join_if
            )
        elif self.m_mode == LEFT_JOIN:
            entries = left_join_if(
                self.m_left_entries,
                self.m_right_entries,
                self.m_join_if
            )
        elif self.m_mode == RIGHT_JOIN:
            entries = right_join_if(
                self.m_left_entries,
                self.m_right_entries,
                self.m_join_if
            )
        elif self.m_mode == FULL_OUTER_JOIN:
            entries = full_outer_join_if(
                self.m_left_entries,
                self.m_right_entries,
                self.m_join_if
            )
        else:
            raise ValueError("Invalid mode %s:" % self.m_mode)
        return self.answer(query, entries)

    @property
    def mode(self) -> int:
        """
        Retrieves the join mode used
        in this :py:class:`JoinIfConnector` instance.

        Returns:
            A value among:

            - :py:data:`INNER_JOIN`
            - :py:data:`LEFT_JOIN`
            - :py:data:`RIGHT_JOIN`
            - :py:data:`FULL_OUTER_JOIN`
        """
        return self.m_mode

    @property
    def left(self) -> Connector:
        """
        Retrieves the left :py:class:`Connector` child
        in this :py:class:`JoinIfConnector` instance.

        Returns:
            The left :py:class:`Connector` child.
        """
        return self.m_left

    @property
    def right(self) -> Connector:
        """
        Retrieves the right :py:class:`Connector` child
        in this :py:class:`JoinIfConnector` instance.

        Returns:
            The left :py:class:`Connector` child.
        """
        return self.m_right

    @property
    def join_if(self) -> callable:
        """
        Retrieves the functor to join entries
        in this :py:class:`JoinIfConnector` instance.

        Returns:
            The left :py:class:`Connector` child.
        """
        return self.m_join_if



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

def is_multiple_key(key: object) -> bool:
    """
    Checks whether a key is a multiple key (i.e., a key involving
    multiple attributes).

    Args:
        key (object): A key. Possible types are ``str`` or ``set``,
            ``frozenset`` or ``list`` of strings.

    Returns:
        ``True`` if and only if ``key`` is multiple,
        ``False`` otherwise.
    """
    return isinstance(key, (set, frozenset, list))

def closure(key: object, fds: dict) -> set:
    """
    Compute the closure of a key given a set of functional dependencies.

    Args:
        key (object): A key. Possible types are ``str`` or ``set``,
            ``frozenset`` or ``list`` of strings.
        fds (dict): A dictionary where each key-value pair represents
            a functional dependency (by mapping a key with another key).

    Returns:
        The "reachable" attributes from ``key``.
    """
    ret = (key | {k for k in key}) if is_multiple_key(key) else {key}
    changed = True

    while changed:
        changed = False
        for (key, column) in fds:
            if key in ret and column not in ret:
                ret.add(column)
                changed = True

    if is_multiple_key(key):
        ret.remove(key)

    return ret

def minimal_cover(fds: set) -> set:
    """
    Retricts a set of functional dependencies to make it 3-NF.

    Args:
        fds (dict): A dictionary where each key-value pair represents
            a functional dependency (by mapping a key with another key).

    Returns:
        The corresponding subset of functional dependencies.
    """
    min_fds = fds.copy()

    for (key, column) in fds:
        # Try to remove (key, column) and see if we cover the closure remains the same
        if column in closure(key, min_fds - {(key, column)}):
            min_fds -= {(key, column)}

    for (key, column) in min_fds.copy():
        if is_multiple_key(key):
            # Check whether each k attribute is required, otherwise, discard it.
            for k in key:
                if k in closure(key - k, min_fds - {(key, column)}):
                    min_fds -= {(key, column)}
                    min_fds |= {(key - k, column)}

    return min_fds

try:
    from pybgl.graph import DirectedGraph, EdgeDescriptor, edges, source, target

    def edge_to_pair(e: EdgeDescriptor, g: DirectedGraph) -> tuple:
        """
        Converts an arc to a pair of vertices.

        Args:
            e (EdgeDescriptor): An arc of the graph.
            g (DirectedGraph): A graph.

        Returns:
            A ``(u, v)`` pair where
            ``u`` is the source of ``e`` in ``g`` and
            ``v`` is the target of ``e`` in ``g``.
        """
        return (source(e, g), target(e, g))

    def minimal_cover_graph(g: DirectedGraph) -> set:
        """
        Computes the minimal cover set of functional dependencies
        given a graph where each vertex corresponding to an attribute
        and each directed arc to a functional dependency.

        Args:
            g (DirectedGraph): A graph.

        Returns:
            The set of 3-NF functional dependencies.
        """
        return minimal_cover({edge_to_pair(e, g) for e in edges(g)})
except ImportError:
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

from .binary_predicate    import BinaryPredicate
from .connector           import Connector
from .query               import Query, ACTION_READ
from .where               import where

try:
    from pymongo import MongoClient
except ImportError as e:
    from .log import Log
    Log.warning(
        "Please install pymongo.\n"
        "  APT: sudo apt install python3-pymongo\n"
        "  PIP: sudo pip3 install --upgrade pymongo\n"
    )
    raise e

class MongoConnector(Connector):
    """
    The :py:class:`MongoConnector` is a minifold gateway allowing
    to manipulate data stored in a Mongo database.
    """
    def __init__(self, mongo_url: str, db_name: str):
        """
        Constructor.

        Args:
            mongo_url (str): The URL of the mongo database.
            db_name (str): The name of the queried database.
        """
        self.client = self.connect(mongo_url)
        self.db_name = db_name
        self.use_database(db_name)

    def connect(self, mongo_url: str) -> MongoClient:
        """
        Connects to a Mongo database.

        Args:
            mongo_url (str): The URL of the mongo database.

        Returns:
            The corresponding :py:class`MongoClient` instance.
        """
        return MongoClient(mongo_url)

    def use_database(self, db_name: str):
        """
        Selects the active database.

        Args:
            db_name (str): The name of the queried database.
        """
        self.db = self.client[db_name]

    def attributes(self, obj: str = None) -> set:
        """
        Lists the available attributes related to a given collection of
        minifold entries exposed by this :py:class:`MongoConnector` instance.

        Args:
            object (str): The name of the collection.

        Returns:
            The set of corresponding attributes.
        """

        return set()

    def query(self, query: Query) -> list:
        """
        Handles an input :py:class:`Query` instance.

        Args:
            query (Query): The handled query.

        Returns:
            The list of entries matching the input query.
        """
        super().query(query)
        ret = list()
        assert self.db
        if query.action == ACTION_READ:
            ret = list(
                self.db[query.object].find(
                    {},
                    {attr:  1 for attr in query.attributes} if query.attributes else None
                )
                    .limit(query.limit if query.limit else 0)
                    .skip(query.offset if query.offset else 0)
            )
            if not query.filters:
                pass
            elif isinstance(query.filters, BinaryPredicate):
                raise RuntimeError("MongoConnector::query: Not yet implemented")
            else:
                return where(ret, query.filters)
        else:
            raise RuntimeError("MongoConnector::query: Unable to query")
        return self.answer(query, ret)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

from .connector import Connector
from .query import Query

def unnest(map_key_unnestedkey: dict, entries: list) -> list:
    """
    Implements the ``unnest`` PostgresSQL function for a list of minifold entries.

    Example:
        >>> unnest({"a": "A", "b": "B"}, [{"a" : [1, 2, 3], "b" : [10, 20, 30]}])
        [{'A': 1}, {'A': 2}, {'A': 3}, {'B': 10}, {'B': 20}, {'B': 30}]

    Args:
        map_key_unnestedkey (dict): A dictionary which maps key corresponding
            to its corresponding unnested key.
        entries (list): A list of minifold entries.

    Returns:
        The list unnested entries. Each entry is a dictionary with a single
        key-value pair, where the key is an unnested key.
    """

    ret = list()
    for entry in entries:
        # NOTE: This only works if entries a single key matching map_key_unnestedkey
        for (k, values) in entry.items():
            new_attribute = map_key_unnestedkey.get(k)
            if isinstance(values, list):
                for value in values:
                    ret.append({new_attribute : value})
            else:
                ret.append({new_attribute : values})
    return ret

class UnnestConnector(Connector):
    """
    The :py:class:`UnnestConnector` class implements the unnest
    function (in PostgreSQL) in a minifold pipeline.
    """
    def __init__(self, map_key_unnestedkey: dict, child: Connector):
        """
        Constructor.

        Args:
            child (Connector): The child minifold :py:class:`Connector` instance.
        """
        super().__init__()
        self.m_child = child
        self.m_map_key_unnestedkey = map_key_unnestedkey

    def attributes(self, object: str) -> set:
        """
        Lists the available attributes related to a given collection of
        minifold entries exposed by this :py:class:`UnnestConnector` instance.

        Args:
            object (str): The name of the collection.

        Returns:
            The set of corresponding attributes.
        """
        raise NotImplementedError

    @property
    def child(self) -> Connector:
        """
        Accessor to the child minifold :py:class:`Connector` instance.

        Returns:
            The child minifold :py:class:`Connector` instance.
        """
        return self.m_child

    def unnest_key(self, key: str) -> str:
        """
        Retrieves the unnested key corresponding to a given key.

        Args:
            key (str): A key identifying an minifold entry
                attribute whose value is a list.

        Raises:
            :py:class:`KeyError` if ``key`` is not in ``self.m_map_key_unnestedkey``.

        Returns:
            The corresponding unnested key.
        """
        return self.m_map_key_unnestedkey[key]

    def query(self, q: Query) -> list:
        """
        Handles an input :py:class:`Query` instance.

        Args:
            query (Query): The handled query.

        Returns:
            The list of entries matching the input query.
        """
        super().query(q)
        return self.answer(
            q,
            unnest(
                self.m_map_key_unnestedkey,
                self.child.query(q)
            )
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

import sys
from pprint import pformat

DEBUG = 0
INFO = 1
WARNING = 2
ERROR = 3

# Shell colors
DEFAULT = 0
RED = 1
GREEN = 2
YELLOW = 3
BLUE = 4
PINK = 5
CYAN = 6
GRAY = 7

# Shell style
BOLD = 1
UNDERLINED = 4
BLINKING = 5
HIGHLIGHTED = 7

class Log:
    """
    The :py:class:`Log` enables logging in Minifold.

    Example:
        >>> Log.enable_print = True
        >>> Log.info("hello")
    """
    enable_print = False  # Set this variable to True to enable logging.
    with_color = True
    log_level = 0

    message_header = {
        DEBUG: "DEBUG",
        INFO: "INFO",
        WARNING: "WARNING",
        ERROR: "ERROR",
    }

    message_color = {
        DEBUG: CYAN,
        INFO: GREEN,
        WARNING: YELLOW,
        ERROR: RED,
    }

    @staticmethod
    def start_style(
        fg_color: int = None,
        bg_color: int = None,
        styles: list = None
    ) -> str:
        """
        Crafts a shell escape sequence to start a style.

        Args:
            fg_color (int): The integer identifying the foreground color to be used.
                _Example:_ ``GREEN``.
            bg_color (int): The integer identifying the background color to be used.
                _Example:_ ``PINK``.
            styles (list) A list of extra styles.
                _Example:_ ``[UNDERLINED, BLINKING]``.

        Returns:
            The corresponding shell string.
        """
        styling = list()
        if fg_color:
            styling.append("3%d" % fg_color)
        if bg_color:
            styling.append("4%d" % bg_color)
        if styles:
            styling += styles
        return "\033[%sm" % ";".join(styling) if styling else ""

    @staticmethod
    def default_style() -> str:
        """
        Crafts a shell escape sequence to set the default style
        """
        return "\033[0m"

    @classmethod
    def print(cls, message_type: int, message: str, file = sys.stderr):
        """
        Internal method, used to orints a message with a custom header and style.

        Args:
            message_type (int): A value in ``INFO``, ``DEBUG``, ``WARNING``, ``ERROR``.
            message (str): The message to log.
            file: The output stream. Defaults to ``sys.stderr``.
        """
        if cls.enable_print and message_type >= cls.log_level:
            color = cls.message_color[message_type]
            header = cls.message_header[message_type]
            print(
                "%(start_style)s%(message)s%(end_style)s" % {
                    "start_style" : cls.start_style(fg_color = color),
                    "message"     : " ".join([
                            header,
                            message if isinstance(message, str) else pformat(message)
                        ]),
                    "end_style"   : cls.default_style()
                },
                file = file
            )

    @classmethod
    def debug(cls, s: str):
        """
        Logs a debug message.

        Args:
            cls (Log): The :py:class:`Log` class.
            s (str): The string to log.
        """
        cls.print(DEBUG, s)

    @classmethod
    def info(cls, s: str):
        """
        Logs an info message.

        Args:
            cls (Log): The :py:class:`Log` class.
            s (str): The string to log.
        """
        cls.print(INFO, s)

    @classmethod
    def warning(cls, s: str):
        """
        Logs a warning message.

        Args:
            cls (Log): The :py:class:`Log` class.
            s (str): The string to log.
        """
        cls.print(WARNING, s)

    @classmethod
    def error(cls, s: str):
        """
        Logs an error message.

        Args:
            cls (Log): The :py:class:`Log` class.
            s (str): The string to log.
        """
        cls.print(ERROR, s)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

import datetime, functools, json, os, pickle, sys, traceback

from functools import partial
from pprint import pprint

from .connector import Connector
from .filesystem import check_writable_directory, mtime, mkdir, rm
from .query import Query, ACTION_READ
from .log import Log


class CacheConnector(Connector):
    """
    :py:class:`CacheConnector` is a :py:class:`Connector` is an
    abstract class used to cache results in the middle of a
    minifold query plan.
    If a query reaches this :py:class:`CacheEntriesConnector` and is
    in cache, it is not forwarded to the underlying connector, hence
    accelerating the query plan execution.

    See specializations:

    - :py:class:`StorageCacheConnector` (caching using a local file)
    - :py:class:`JsonCacheConnector` (caching using a local JSON file)
    - :py:class:`PickleCacheConnector` (caching using a local pickle file)

    Possible improvements:

    - For the moment the cache is class-name based. It should rather
      be identify by the connector setup and the underlying connectors
      (if any)
    - For the moment, the cache is only used if the exact same Query
      was issued over the past. But we should be able to reuse the cache
      if a less strict query has been issued, and use
      Connector.reshape_entries afterwards.
    """
    def __init__(self, child):
        """
        Constructor.
        """
        self.child = child

    def attributes(self, object: str) -> set:
        """
        Retrieves the keys supported by this
        :py:class:`CacheEntriesConnector` instance.

        Returns:
            The key of the underlying entries.
        """
        return self.child.attributes(object)

    def callback_read(self, query: Query) -> object:
        """
        Callback triggered when data must be fetched in this
        :py:class:`CacheEntriesConnector`.

        Args:
            query (Query): The handled :py:class:`Query` instance.

        Raise:
            RuntimeError: if not overloaded.

        Returns:
            The fetched data.
        """
        raise RuntimeError("Must be overloaded")

    def callback_write(self, query: Query, data: object):
        """
        Callback triggered when data must be saved in this
        :py:class:`CacheEntriesConnector` instance.

        Args:
            query (Query): The handled :py:class:`Query` instance.
            data (object): The data to be saved.

        Raise:
            RuntimeError: if not overloaded.
        """
        raise RuntimeError("Must be overloaded")

    def clear_query(self, query: Query):
        """
        Removes a :py:class:`Query` result from this
        :py:class:`CacheEntriesConnector` instance.
        This method should be overloaded.

        Args:
            query (Query): The handled :py:class:`Query` instance.
        """
        pass

    def clear_cache(self):
        """
        Clears this :py:class:`CacheEntriesConnector` instance entirely.
        This method should be overloaded.
        """
        pass

    def read(self, query: Query) -> object:
        """
        Fetches from this :py:class:`CacheEntriesConnector` instance the corresponding data.

        Returns:
            The corresponding cached object.
        """
        (data, success) = (None, False)
        try:
            data = self.callback_read(query)
            success = (data is not None)
        except:
            Log.error(
                "CacheConnector.read(%s): Cannot read cache:\n%s" % (
                    query,
                    traceback.format_exc()
                )
            )
        return (data, success)

    def write(self, query: Query, data: object) -> bool:
        """
        Writes data to this :py:class:`CacheEntriesConnector` instance.

        Args:
            query (Query): The handled :py:class:`Query` instance.
            data (object): The data fetched by this :py:class:`Query` that must be
                saved to this :py:class:`CacheEntriesConnector` instance.

        Returns:
            The corresponding cached object.
        """
        success = True
        try:
            self.callback_write(query, data)
        except:
            Log.error(
                "CacheConnector.write(%s, %s): Cannot write cache:\n%s" % (
                    traceback.format_exc(),
                    query,
                    data
                )
            )
            success = False
        return success

    def is_cached(self, query: Query) -> bool:
        """
        Checks whether a :py:class:`Query` instance is already cached in this
        :py:class:`CacheEntriesConnector` instance.

        Args:
            query (Query): The handled :py:class:`Query` instance.

        Returns:
            ``True`` if ``query`` is cached in this
            :py:class:`CacheEntriesConnector` instance,
            ``False`` otherwise.
        """
        raise RuntimeError("Must be overloaded")

    def is_cachable(self, query: Query, data: object) -> bool:
        """
        Checks whether a data may be saved in this
        :py:class:`CacheEntriesConnector` instance.

        Args:
            query (Query): The handled :py:class:`Query` instance.
            data (object): The data fetched by this :py:class:`Query` that must be
                saved to this :py:class:`CacheEntriesConnector` instance.

        Returns:
            ``True`` if ``query`` may be cached in this
            :py:class:`CacheEntriesConnector` instance,
            ``False`` otherwise.
        """
        return True

    def query(self, query: Query) -> list:
        """
        Handles an incoming :py:class:`Query` instance.

        The :py:class:`CacheEntriesConnector` checks whether it is already cached.
        If ``query`` is cached in this :py:class:`CacheEntriesConnector`, it is not
        forwarded to :py:attr:`self.child` and the results are directly from the cache.
        Otherwise, it is forwarded to :py:attr:`self.child`.

        Args:
            query (Query): The handled :py:class:`Query` instance.

        Returns:
            The corresponding entries.
        """
        (data, success) = (None, False)
        if self.is_cached(query):
            (data, success) = self.read(query)
            if not success:
                Log.warning("CacheConnector.query(%s): Unreadable cache" % query)
        if not success:
            data = self.child.query(query)
        if query.action == ACTION_READ and self.is_cachable(query, data):
            self.write(query, data)
        return self.answer(query, data)


# Default parameters, used to initialize StorageCacheConnector class members.
DEFAULT_CACHE_STORAGE_BASE_DIR = os.path.join(os.path.expanduser("~"), ".minifold", "cache")
DEFAULT_CACHE_STORAGE_LIFETIME = datetime.timedelta(days=3)


def make_cache_dir(base_dir: str, sub_dir: str = ""):
    """
    Crafts the path to a minifold cache file.

    Args:
        base_dir (str): The directory storing the minifold caches.
            See also :py:data:`DEFAULT_CACHE_STORAGE_BASE_DIR`.
        sub_dir (str): The directory storing the minifold caches.
            This is often the name assigned to a connector pulling
            remote data.
    """
    return os.path.join(base_dir, sub_dir) if sub_dir else base_dir


class StorageCacheConnector(CacheConnector):
    """
    :py:class:`StorageCacheConnector` is an abstract class
    that specializes of :py:class:`CacheConnector` to manage file on the local storage.

    See specializations:

    - :py:class:`JsonCacheConnector` (caching using a local JSON file)
    - :py:class:`PickleCacheConnector` (caching using a local pickle file)
    """
    base_dir = DEFAULT_CACHE_STORAGE_BASE_DIR  # Path to the minifold cache directory
    lifetime = DEFAULT_CACHE_STORAGE_LIFETIME  # Lifetime of the cached objects

    def __init__(
            self,
            child: Connector,
            callback_load: callable = None,
            callback_dump: callable = None,
            lifetime: datetime.timedelta = None,
            cache_dir: str = None,
            read_mode: str = "r",
            write_mode: str = "w",
            extension: str = ""
    ):
        """
        Constructor.

        Args:
            child (Connector): The child :py:class:`Connector` instance.
            callback_load (callable): A function ``callback_load(f)`` where
                ``f`` is the read file descriptor of the cache and which returns
                the cached object.
            callback_dump (callable): A function ``callback_load(data, f)`` where
                ``f`` is the read file descriptor of the cache
                and ``data`` is the object to be cached.
            lifetime (datetime.timedelta): The lifetime of the cached objects.
                Pass ``None`` to use the default lifetime.
            cache_dir (str): The path to the cache directory.
                Pass ``None`` to use the default minifold cache directory.
            read_mode (str): The string specifying how to open the file descriptor
                to read the cache. Possible values are ``"r"`` (text cache)
                and ``"rb"`` (binary cache).
            read_mode (str): The string specifying how to open the file descriptor
                to read the cache. Possible values are ``"r"`` (text cache)
                and ``"rb"`` (binary cache).
            extension (str): The extension of the cache filename.
        """
        super().__init__(child)
        self.callback_load = callback_load
        self.callback_dump = callback_dump
        self.lifetime = lifetime if lifetime is not None else StorageCacheConnector.lifetime
        self.cache_dir = cache_dir if cache_dir else \
            make_cache_dir(StorageCacheConnector.base_dir, child.__class__.__name__)
        self.read_mode = read_mode
        self.write_mode = write_mode
        self.extension = extension

    def make_cache_filename(self, query: Query) -> str:
        """
        Crafts the filename of the cache to store a given :py:class:`Query` instance.

        Args:
            query (Query): The handled :py:class:`Query` instance.

        Returns:
            The corresponding filename.
        """
        return os.path.join(self.cache_dir, str(query) + self.extension)

    def clear_query(self, query: Query):
        """
        Removes a :py:class:`Query` result from this
        :py:class:`StorageCacheConnector` instance.

        Args:
            query (Query): The handled :py:class:`Query` instance.
        """
        cache_filename = self.make_cache_filename(query)
        if os.path.exists(cache_filename):
            Log.debug("StorageCacheConnector: Removing query [%s]" % cache_filename)
            rm(cache_filename)

    def clear_cache(self):
        """
        Clears this :py:class:`CacheEntriesConnector` instance entirely.
        """
        if os.path.exists(self.cache_dir) and os.path.isdir(self.cache_dir):
            Log.debug("StorageCacheConnector: Removing cache [%s]" % self.cache_dir)
            rm(self.cache_dir, recursive=True)

    @staticmethod
    def is_fresh_cache(cache_filename: str, lifetime: datetime.timedelta) -> bool:
        """
        Checks whether a minifold cache file is fresh enough to be relevant.

        To do so, :py:class:`StorageCacheConnector.is_fresh_cache` fetches the
        date of the cache file from the filesystem and matches it against
        ``lifetime``.

        Args:
            cache_filename (str): The path to the cache.
            lifetime (datetime.timedelta): The lifetime of the cache.

        Returns:
            ``True`` if and only if the cache is fresh enough,
            ``False`` otherwise.
        """
        is_fresh = True
        if lifetime:
            t_now = datetime.datetime.utcnow()
            t_cache = mtime(cache_filename)
            is_fresh = (t_now - t_cache) < lifetime
            # Log.debug("t_now(%s) - t_cache(%s) = %s ?< lifetime %s" % (
            #    t_now, t_cache, (t_now - t_cache), lifetime
            # ))
        return is_fresh

    def is_cached(self, query: Query) -> bool:
        """
        Checks whether a :py:class:`Query` instance is already cached in this
        :py:class:`StorageCacheConnector` instance.

        Args:
            query (Query): The handled :py:class:`Query` instance.

        Returns:
            ``True`` if ``query`` is cached in this
            :py:class:`StorageCacheConnector` instance,
            ``False`` otherwise.
        """
        ret = False
        cache_filename = self.make_cache_filename(query)
        if os.path.exists(cache_filename):
            ret = StorageCacheConnector.is_fresh_cache(cache_filename, self.lifetime)
        return ret

    def callback_read(self, query: Query) -> object:
        """
        Callback triggered when data must be fetched in this
        :py:class:`StorageCacheConnector`.

        Args:
            query (Query): The handled :py:class:`Query` instance.

        Returns:
            The fetched data.
        """
        data = None
        cache_filename = self.make_cache_filename(query)
        if self.is_cached(query):
            with open(cache_filename, self.read_mode) as f:
                data = self.callback_load(f)
            Log.debug("Cache hit: [%s]" % cache_filename)
        return data

    def callback_write(self, query: Query, data: object):
        """
        Callback triggered when data must be saved in this
        :py:class:`StorageCacheConnector` instance.

        Args:
            query (Query): The handled :py:class:`Query` instance.
            data (object): The data to be saved.
        """
        cache_filename = self.make_cache_filename(query)
        directory = os.path.dirname(cache_filename)
        mkdir(directory)
        check_writable_directory(directory)
        with open(cache_filename, self.write_mode) as f:
            self.callback_dump(data, f)


class PickleCacheConnector(StorageCacheConnector):
    """
    :py:class:`PickleCacheConnector` overloads :py:class:`StorageCacheConnector`
    to cache result in pickle files.
    """
    def __init__(
            self,
            child: Connector,
            lifetime: datetime.timedelta = StorageCacheConnector.lifetime,
            cache_dir: str = None
    ):
        """
        Constructor.

        Args:
            child (Connector): The child :py:class:`Connector` instance.
            lifetime (datetime.timedelta): The lifetime of the cached objects.
                Pass ``None`` to use the default lifetime.
            cache_dir (str): The path to the cache directory.
                Pass ``None`` to use the default minifold cache directory.
        """
        super().__init__(
            child,
            pickle.load, pickle.dump,
            lifetime, cache_dir,
            "rb", "wb", ".pkl"
        )


class JsonCacheConnector(StorageCacheConnector):
    """
    :py:class:`JsonCacheConnector` overloads :py:class:`StorageCacheConnector`
    to cache result in JSON files.
    """
    def __init__(
            self,
            child: Connector,
            lifetime: datetime.timedelta = StorageCacheConnector.lifetime,
            cache_dir: str = None
    ):
        """
        Constructor.

        Args:
            child (Connector): The child :py:class:`Connector` instance.
            lifetime (datetime.timedelta): The lifetime of the cached objects.
                Pass ``None`` to use the default lifetime.
            cache_dir (str): The path to the cache directory.
                Pass ``None`` to use the default minifold cache directory.
        """
        super().__init__(
            child,
            json.load, partial(json.dump, indent=4),
            lifetime, cache_dir,
            "r", "w", ".json"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

import re
from .where import where

def equals(x: object, y: object) -> bool:
    """
    Tests whether two objects are equal (wraps the ``==`` operator).

    Examples:
        >>> equals("foo", "foo")
        True
        >>> equals("foo", "FOO")
        False

    Args:
        x (object): An object.
        y (object): An object.

    Returns:
        ``True`` if and only if ``x == y``,
        ``False`` otherwise.
    """
    return x == y

def contains(x: object, y: object) -> bool:
    """
    Tests whether two objects are equal (wraps the ``==`` operator).

    Example:
        >>> equals("bar", "foobar")
        True

    Args:
        x (object): An object.
        y (object): An object.

    Returns:
        ``True`` if and only if ``x in y``,
        ``False`` otherwise.
    """
    return x in y

def lower_case_equals(x: str, y: str) -> bool:
    """
    Checks whether two string are equals without considering the case.

    Example:
        >>> lower_case_equals("foo", "FOO")
        True

    Args:
        x (str): A string.
        y (str): A string.

    Returns:
        ``True`` if and only if the both strings are equal without considering the case,
        ``False`` otherwise.
    """
    return x.lower() == y.lower()

def lower_case_contains(x: str, y: str) -> bool:
    """
    Checks whether a string is included in another one without considering the case.

    Example:
        >>> lower_case_contains("foo", "barFOObar")
        True

    Args:
        x (str): The search string.
        y (str): The queried string.

    Returns:
        ``True`` if and only if the both strings are equal without considering the case,
        ``False`` otherwise.
    """
    return x.lower() in y.lower()

def contains_words(word: str, sentence: str, ignore_case: bool = True):
    """
    Checks whether a word is contained in a string (e.g., a sentence).

    Examples:
        >>> contains_words("earth", "Earth is a planet.")
        True
        >>> contains_words("earth", "A terrible earthquake.")
        False

    Args:
        word (str): The searched word.
        sentence (str): The queried string.
        ignore_case (bool): Pass `True` if the search is not case sensitive,
            ``False`` otherwise. Defaults to ``True``.

    Returns:
        ``True`` if ``word`` has been found in ``sentence``,
        ``False`` otherwise.
    """
    pattern = ".*\\b(%s)\\b.*" % word
    flags = re.IGNORECASE if ignore_case else 0
    regex = re.compile(pattern, flags)
    return bool(regex.match(sentence))

class SearchFilter:
    """
    The :py:class:`SearchFilter` implements a minifold filter allowing
    to filter minifold entries based on a search predicate.
    """
    def __init__(self, search_values: list, attributes: list, match: callable):
        """
        Constructor.

        Args:
            search_values (list): The searched values.
            attributes (list): The attributes of interest.
            match (callable): The search strategy, e.g.,
                :py:func:`equals`,
                :py:func:`contains`.
                If the values are string, you could also consider:
                :py:func:`lower_case_contains`,
                :py:func:`lower_case_equals`,
                :py:func:`contains_words`.
        """
        self.search_values = search_values
        self.attributes = attributes
        self.match = match

    def __call__(self, entry: dict) -> bool:
        """
        Functor method.

        Args:
            entry (dict): A minifold entry.

        Returns:
            ``True`` if ``entry`` is matched by the search,
            ``False`` otherwise.
        """
        for search_value in self.search_values:
            for attribute in self.attributes:
                if self.match(search_value, entry[attribute]):
                    return True
        return False

def search(
    entries: list,
    attributes: list,
    search_values: list,
    match: callable = equals
) -> list:
    """
    Searches the entries whose at least one attribute of interest is matched
    a searched word

    Args:
        entries (list): A list of minifold entries.
        attributes (list): The attributes of interest.
        search_values (list): The searched values.
        match (callable): The search strategy, e.g.,
            :py:func:`equals`,
            :py:func:`contains`.
            If the values are string, you could also consider:
            :py:func:`lower_case_contains`,
            :py:func:`lower_case_equals`,
            :py:func:`contains_words`.

    Returns:
        The subset of entries matched by the search.
    """
    return where(entries, SearchFilter(search_values, attributes, match))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

"""
The DBLP Computer science bibliography provides open bibliographic information
on major computer science journals and proceedings.

- `DBLP website <https://dblp.org/>`__
- `API specs <https://wiki.inria.fr/lincs/Dblp-api>`__

Some query examples:

- `Retrieves articles authored by Fabien Mathieu, in JSON format <https://dblp.dagstuhl.de/search/publ/api?q=fabien-mathieu&h=500&format=json>`
- `Retrieves articles authored by Fabien Mathieu pubished in 2014, in JSON format <https://dblp.dagstuhl.de/search/publ/api?q=fabien-mathieu%20year:2014&h=500&format=json>`

By default DBLP only returns up to 30 records (see
`this link <https://dblp.org/faq/How+to+use+the+dblp+search+API.html>`__).

The default limit in :py:class:`DblpConnector` is set to ``9999``.

It is possible to query a specific researcher using its DBLP-ID.
The ID can be found by browsing the page related to a researcher.

Example:
- `Query using the DBLP name <https://dblp.uni-trier.de/pers/hd/c/Chen:Chung_Shue>`
- `Query using the DBLP ID <https://dblp.org/pid/30/1446>`
- The DBLP ID can be obtained by clicking on the export bibliography icon.

For the moment, the only result format supported by DBLP is XML
(see `this link <https://dblp.org/pid/30/1446.xml>`__).
"""

try:
    import urllib3
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
except ImportError as e:
    from .log import Log
    Log.warning(
        "\n".join([
            "Please install requests",
            "  APT: sudo apt install python3-urllib3",
            "  PIP: sudo pip3 install --upgrade urllib3",
        ])
    )
    raise e

try:
    import requests
except ImportError as e:
    from .log import Log
    Log.warning(
        "\n".join([
            "Please install requests",
            "  APT: sudo apt install python3-requests",
            "  PIP: sudo pip3 install --upgrade requests",
        ])
    )
    raise e

try:
    import xmltodict
except ImportError as e:
    from .log import Log
    Log.warning(
        "\n".join([
            "Please install requests",
            "  APT: sudo apt install python3-xmltodict",
            "  PIP: sudo pip3 install --upgrade xmltodict",
        ])
    )
    raise e

import json
import operator
from pprint import pformat

from .binary_predicate import BinaryPredicate
from .connector import Connector
from .doc_type import DocType
from .log import Log
from .strings import to_international_string, to_canonic_fullname as _to_canonic_fullname
from .query import Query, ACTION_READ

# Default queried DBLP API.
DBLP_API_URL = "https://dblp.dagstuhl.de"

# Maps DBLP ontology to our ontology.
DBLP_ALIASES = {
    "type"    : "dblp_doc_type",
    "venue"   : "conference",
    "url"     : "dblp_url",
}

def to_canonic_fullname(s: str) -> str:
    """
    Canonizes an author full name.

    Args:
        s (str): The input fullname.

    Returns:
        The canonizes fullname.
    """
    # DBLP may now returns {"@pid" : int, "text" : "Firstname Lastname"}
    # instead of the author fullname.
    if isinstance(s, dict):
        s = s["text"]
    return _to_canonic_fullname(s)

class DblpConnector(Connector):
    """
    The :py:class:`DblpConnector` class is a minifold gateway allowing to
    fetch data from DBLP (repository of scientific articles).

    See also:
    - :py:class:`HalConnector`.
    - :py:class:`GoogleScholarConnector`.
    """
    def __init__(
        self,
        map_dblp_id: dict = None,
        map_dblp_name: dict = None,
        dblp_api_url: str = DBLP_API_URL
    ):
        """
        Constructor.

        Args:
            map_dblp_id (dict): Maps an author full name with his/her DBLP ID.
            map_dblp_name (dict): Maps an author obsolete full name with his/her
                current full name.
            dblp_api_url (str): The URL of the DBLP server.
                Defaults to :py:data:`DBLP_API_URL`.
        """
        super().__init__()
        self.m_api_url = dblp_api_url
        assert (
            not dblp_api_url.endswith("/search") and
            not dblp_api_url.endswith("/search/")
        ), f"Invalid API URL [{dblp_api_url}], remove '/search/' suffix"
        self.m_format  = "json"  # valid values are in {"xml", "json", "jsonp"}.
        self.m_map_dblp_id = map_dblp_id if map_dblp_id else dict()
        self.m_map_dblp_name = map_dblp_name if map_dblp_name else dict()
        self.m_map_rev_name = {
            to_canonic_fullname(dblp_name) : name
            for (name, dblp_name) in self.m_map_dblp_name.items()
        }

    def attributes(self, object: str) -> set:
        """
        Lists the attributes of the collection of objects stored
        in this :py:class:`DblpConnector` instance.

        Args:
            object (str): The name of the minifold object.
                As a :py:class:`DblpConnector` instance stores a single
                collection, ``object`` is no relevant and you may pass ``None``.

        Returns:
            The set of available ``object``'s attributes
        """

        return {
            "authors", "doc_type", "title", "type", "venue", "url", "year"
        }  # Non exhaustive

    @property
    def api_url(self) -> str:
        """
        Retrieves the URL of the remote DBLP server. managed by
        this :py:class:`DblpConnector` instance.

        Returns:
            The URL of the remote DBLP server.
        """
        return self.m_api_url

    @property
    def format(self) -> str:
        """
        Retrieves the format of data retrieved from the DBLP server.

        Returns:
            A value in ``{"xml", "json", "jsonp"}``.
        """
        return self.m_format

    @property
    def map_dblp_name(self) -> dict:
        """
        Retrieves the dictionary that maps an author obsolete full name with his/her
        current full name.

        Returns:
            The queried dictionary.
        """
        return self.m_map_dblp_name

    @property
    def map_dblp_id(self) -> dict:
        """
        Retrieves the dictionary that maps an author full name with his/her DBLP ID.

        Returns:
            The queried dictionary.
        """
        return self.m_map_dblp_id

    @staticmethod
    def to_dblp_name(s: str, map_dblp_name: dict = None) -> str:
        """
        Converts a fullname to the corresponding current DBLP name.

        Args:
            s (str): The input fullname.

        Returns:
            The (possibly updated) fullname.
        """
        if map_dblp_name:
            s = map_dblp_name.get(s, s)
        s = to_international_string(s)
        words = s.lower().split()
        try:
            words.remove("")
        except ValueError:
            pass
        # a-b means that b must follows a
        # a_b means that both a and b must be found
        # $ corresponds to exact match
        # See: http://dblp.uni-trier.de/db/about/author.html
        ret = "-".join(words) + "$"
        return ret

    def get_dblp_name(self, s: str) -> str:
        """
        Retrieves the current DBLP fullname of an author.

        Args:
            s (str): The input fullname.

        Returns:
            The (possibly updated) fullname.
        """
        return DblpConnector.to_dblp_name(s, self.map_dblp_name)

    def get_dblp_id(self, s: str) -> str:
        """
        Retrieves the current DBLP ID of an author.

        Args:
            s (str): The input fullname.

        Returns:
            The corresponding DBLP ID or ``s``.
        """
        pid = self.m_map_dblp_id.get(s)
        return pid if pid else self.get_dblp_name(s)

    def binary_predicate_to_dblp(self, p :BinaryPredicate, result :dict):
        """
        Converts a minifold predicate to a DBLP predicate.

        Args:
            p (BinaryPredicate): A :py:class:`BinaryPredicate` instance.
            result (dict): The output dictionary.
        """
        # Recursive call only supported for && clauses
        if p.operator == operator.__and__:
            self.binary_predicate_to_dblp(p.left, result),
            self.binary_predicate_to_dblp(p.right, result)
            return

        # Simple predicate. The left member must be a attribute name of the entry.
        if not isinstance(p.left, str):
            raise RuntimeError(
                f"binary_predicate_to_dblp: the left operand of {p} must be a string"
            )

        if p.left in ["author", "authors", "researcher", "conference"]:
            # String attribute. Only "CONTAINS" and "==" are supported.
            # "==" behaves like "CONTAINS"
            if p.operator == operator.__eq__ or p.operator == operator.__contains__:
                result["prefix"] = self.get_dblp_name(p.right)
            else:
                raise RuntimeError(
                    f"binary_predicate_to_dblp: unsupported operator ({p.operator}): {p}"
                )
        else:
            # Other attributes. Only "==" is supported.
            if p.operator == operator.__eq__:
                result["suffix"] += "%20" + ("%s:%s" % (p.left, p.right))
            else:
                raise RuntimeError(
                    f"binary_predicate_to_dblp: unsupported operator ({p.operator}): {p}"
                )

    @staticmethod
    def to_doc_type(s: str) -> DocType:
        """
        Converts a DBLP doc type to the corresponding :py:class:`DocType` instance.

        Args:
            s (str): The DBLP doc type.

        Returns:
            The corresponding :py:class:`DocType` instance.
        """
        s = s.lower()
        if s in {
            "article", "conference and workshop papers", "conference or workshop",
            "incollection", "inproceedings", "proceedings"
        }:
            return DocType.ARTICLE
        # elif s == "????":
        #     return DocType.COMM
        elif s == "journal articles":
            return DocType.JOURNAL
        elif s in {"informal publications", "reference works"}:
            return DocType.REPORT
        elif s in {"phdthesis", "books and theses", "parts in books or collections"}:
            return DocType.BOOKS_AND_THESES
        # elif s == "????":
        #     return DocType.HDR
        elif s == "editorship":
            return DocType.CHAPTER
        else:
            Log.warning("DblpConnector.to_doc_type: unknown type: %s" % s)
            return DocType.UNKNOWN

    def reshape_entry(self, query: Query, entry: dict) -> dict:
        """
        Reshape a DBLP entry to make it compliant with a given
        :py:class:`Query` instance.

        Args:
            query (Query): A :py:class:`Query` instance.
            entry (dict): The entry to reshape.

        Returns:
            The reshaped entry.
        """
        if "type" in entry.keys():
            doc_type = DblpConnector.to_doc_type(entry["type"])
            entry["doc_type"]      = doc_type  # Compatible with Hal "doc_type" values.
            entry["dblp_doc_type"] = doc_type  # To compare Hal and DBLP doc_types.

        if len(query.attributes) > 0:
            keys = set(entry.keys()) & set(query.attributes)
            entry = {k : entry[k] for k in keys}

        for (k, v) in entry.items():
            if isinstance(v, str):
                try:
                    entry[k] = int(v)
                except ValueError:
                    pass

        if "authors" in entry.keys():
            if isinstance(entry["authors"], (dict, str)):
                entry["authors"] = [entry["authors"]]

            # Fix author names having homonyms are not well-named (e.g
            # "Giovanni Pau 0001" instead of "Giovanni Pau").
            entry["authors"] = [
                to_canonic_fullname(author).rstrip(" 0123456789") for author in entry["authors"]
            ]

            # Convert DBLP names to our names if needed.
            entry["authors"] = [
                self.m_map_rev_name.get(_to_canonic_fullname(author), author)
                for author in entry["authors"]
            ]

        return entry

    def extract_entries(self, query: Query, results: list) -> list:
        """
        Extracts the minifold entries for a DBLP query result.

        Args:
            query (Query): A :py:class:`Query` instance.
            results: The DBLP results.

        Returns:
            A ``list`` of ``dict`` instances (the minifold entres).
        """
        entries = list()
        # If the query is not a standard Dblp object, we are pulling a bibliography
        # of a researcher who is identified by query.object (PID or fullname)
        is_bib_query = query.object not in {"publication", "researcher", "conference"}
        canonic_fullname = to_canonic_fullname(self.get_dblp_name(query.object)).rstrip("$") if is_bib_query else None
        try:
            raw_entries = results["result"]["hits"]["hit"]
            for raw_entry in raw_entries:
                entry               = raw_entry["info"]
                entry["dblp_id"]    = int(raw_entry["@id"])
                entry["dblp_score"] = int(raw_entry["@score"])

                try:
                    entry["authors"] = entry.pop("authors")["author"]
                except KeyError:
                    pass

                if is_bib_query:
                    # Unfortunately, DBLP seems unable to perform exact match search.
                    # For example searching "francois-durand$" returns "Jean-François
                    # Durand" publications.
                    # The following discards the false positives.
                    if isinstance(entry["authors"], str):
                        entry["authors"] = [entry["authors"]]

                    if canonic_fullname not in [
                        to_canonic_fullname(author) for author in entry["authors"]
                    ]:
                        # Log.warning("Ignoring %(title)s %(authors)s (homonym)" % entry)
                        continue

                entries.append(entry)
        except KeyError:  # Occurs if 0 DBLP publication found
            pass
        return entries

    def reshape_entries(self, query: Query, entries: list) -> list:
        """
        Apply :py:meth:`DblpConnector.reshape_entry` to a list of entries.

        Args:
            query (Query): A :py:class:`Query` instance.
            entries (list): A list of minifold entries.

        Returns:
            The reshaped entries.
        """
        return [self.reshape_entry(query, entry) for entry in entries]

    def query(self, query: Query) -> list:
        """
        Handles an input :py:class:`Query` instance.

        Args:
            query (Query): The handled query.

        Returns:
            The list of entries matching ``query``.
        """
        super().query(query)
        entries = list()
        if query.action == ACTION_READ:

            pid = None
            format = self.format
            object = ""
            url_options = list()
            if query.object == "publication":
                object = "search/publ"
            elif query.object == "researcher":
                object = "search/author"
            elif query.object == "conference":
                object = "search/venue"
            else:
                fullname = query.object
                pid = self.map_dblp_id.get(fullname)
                dblp_name = self.get_dblp_name(fullname)
                object = "pid" if pid else "search/publ"
                # For the moment, DBLP only supports XML for pid-based queries.
                # https://dblp.org/pid/30/1446.xml
                if pid:
                    format = "xml"
                else:
                    url_options.append(dblp_name)

            if object == "pid":
                q_dblp = "%(server)s/%(object)s/%(pid)s.%(format)s" % {
                    "server" : self.api_url,
                    "object" : object,
                    "pid"    : pid,
                    "format" : format,
                }
            else:
                # WHERE
                if query.filters:
                    search = {
                        "prefix" : self.get_dblp_name(query.object),
                        "suffix" : ""
                    }
                    self.binary_predicate_to_dblp(query.filters, search)
                    url_options.append("%s%s" % (search["prefix"], search["suffix"]))

                # OFFSET and LIMIT
                url_options.append("h=%s" % query.limit if query.limit is not None else "h=9999")
                if query.offset:
                    url_options.append("f=%s" % query.offset)

                # Format of the result.
                url_options.append("format=%s" % format)
                q_dblp = "%(server)s/%(object)s/api?q=%(query)s" % {
                    "server" : self.api_url,
                    "object" : object,
                    "query"  : "&".join(url_options)
                }

            Log.info("--> DBLP: %s" % q_dblp)
            # http = urllib3.PoolManager(timeout=urllib3.Timeout(connect=1.0, read=2.0))
            # reply = http.request("GET", q_dblp, retries = 5)
            reply = requests.get(q_dblp, timeout=(5, 5))

            if reply.status_code == 200:
                data = reply.content.decode("utf-8")
                if format == "json":
                    result = json.loads(data)
                    entries = self.extract_entries(query, result)
                elif format == "xml":
                    data = data.replace("<i>", "")
                    data = data.replace("</i>", "")
                    result = xmltodict.parse(data, dict_constructor=dict)
                    # N.B. There are two other keys of interests
                    # - "co" : coauthors
                    # - "person" : information about the researcher

                    def xml_to_entry(d: dict) -> dict:
                        """
                        Converts a dictionary obtained from XML result returned
                        by DBLP to a flat dictionary.

                        Args:
                            d: The DBLP dictionary.

                        Returns:
                            The corresponding flat dictionary.
                        """
                        publication_type = next(iter(d.keys()))
                        entry = d[publication_type]
                        entry["type"] = publication_type
                        key = \
                            "author" if "author" in entry.keys() else \
                            "editor" if "editor" in entry.keys() else \
                            None
                        if key:
                            # Sometimes, author is represented by a dict with key
                            # '@orcid' and '#text'
                            if isinstance(entry[key], str):
                                entry[key] = [entry[key]]

                            entry["authors"] = [
                                author["#text"] if isinstance(author, dict) else author
                                for author in entry[key]
                            ]
                        else:
                            Log.warning(f"No author found for this DBLP publication:\n{pformat(entry)}")
                        return entry

                    entries = [xml_to_entry(d) for d in result["dblpperson"]["r"]]
                else:
                    raise RuntimeError("Format not implemented: %s" % self.format)
            else:
                raise RuntimeError("Cannot get reply from %s" % self.api_url)

        return self.answer(query, self.reshape_entries(query, entries))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

"""
This file gathers utilities related to ipython Jupyter notebooks.
"""

def in_ipynb() -> bool:
    """
    Tests whether the code is running inside a Jupyter Notebook.

    Returns:
        ``True`` iff the code is running inside a Jupyter Notebook.
    """
    try:
        from IPython import get_ipython
        return str(type(get_ipython())) == "<class 'ipykernel.zmqshell.ZMQInteractiveShell'>"
    except NameError:
        return False
    except ImportError:
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

import json
from .connector import Connector
from .singleton import Singleton

# Example of configuration

DEFAULT_MINIFOLD_CONFIG = """{
    "dblp:dagstuhl" : {
        "type" : "minifold.dblp.DblpConnector",
        "args" : {
            "dblp_api_url" : "https://dblp.dagstuhl.de"
        }
    },
    "dblp:uni-trier" : {
        "type" : "minifold.dblp.DblpConnector",
        "args" : {
            "dblp_api_url" : "https://dblp.uni-trier.de"
        }
    }
}"""


class Config(dict, metaclass=Singleton):
    """
    :py:class:`Config` is used to centralize the minifold
    configuration, possibly fetched from multiple files.

    Note this class inherits :py:class:`Singleton`, meaning
    that any instance of this class corresponds to the same
    instance.

    Each piece of configuration must be a dictionary mapping abitrary,
    where keys are (minifold gateway) arbitary names.
    Each gateway is mapped with a dictionnary which
    maps ``"key"`` with the appropriate :py:class:`Connector` type
    and ``"args"`` with the parameters to be passed to the corresponding constructor.

    These configuration pieces are typically stored in configuration
    file that resides in ``~/.minifold/conf``. By convention, each JSON
    file stored in this directory is named "gw_type:gw_name.json"
    where gw_type helps to understand the underlying minifold connector
    and gw_name identifies the nature of the data source.

    See :py:data:`DEFAULT_MINIFOLD_CONFIG`.

    Example:
        >>> from minifold.dblp import DblpConnector
        >>> config = Config()
        >>> config.loads(DEFAULT_MINIFOLD_CONFIG)
        >>> dblp1 = config.make_connector("dblp:dagstuhl")
        >>> dblp2 = config.make_connector("dblp:uni-trier")
    """
    def loads(self, s_json: str):
        """
        Populates the :py:class:`Config` instance from a JSON string.

        Args:
            s_json (str): A string containing JSON configuration.
        """
        self.update(json.loads(s_json))

    def load(self, stream):
        """
        Populates the :py:class:`Config` instance from an input stream
        (e.g., a read file descriptor) storing JSON data.

        Args:
            stream: The input JSON stream.
        """
        self.update(json.load(stream))

    def load_file(self, filename: str):
        """
        Populates the :py:class:`Config` instance from an JSON file.

        Args:
            filename (str): The path to the input JSON file.
                Example: ``~/.minifold/conf/gw_type:gw_name.json``.
        """
        with open(filename, "r") as f:
            self.load(f)

    def make_connector(self, name: str) -> Connector:
        """
        Makes a :py:class:`Connector` thanks to the :py:class:`Config` instance.

        Args:
            name (str): The name identifying the :py:class:`Connector` in
                the minifold configuration

        Raises:
            KeyError, if ``name`` is not in the :py:class:`Config` instance.

        Returns:
            The corresponding :py:class:`Connector` instance.
        """
        conf = self.get(name)
        if not conf:
            raise KeyError(
                "Config: Key [%s] not found. Known configuration are:\n%s\n" % (
                    name,
                    "\n\t".join([str(k) for k in self.keys()])
                )
            )
        type_name = conf["type"]
        cls = Connector.subclasses.get(type_name)
        if not cls:
            raise KeyError(
                "Config: Connector [%s] not found. Known connectors are:\n%s\n" % (
                    type_name,
                    "\n\t".join([str(k) for k in Connector.subclasses.keys()])
                )
            )
        kwargs = conf.get("args", dict())
        return cls(**kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

from .log       import Log
from .singleton import Singleton

try:
    import requests
except ImportError as e:
    from .log import Log
    Log.warning(
        "Please install requests.\n"
        "  APT: sudo apt install python3-requests\n"
        "  PIP: sudo pip3 install --upgrade requests\n"
    )
    raise e

class Proxy(dict, metaclass=Singleton):
    """
    The :py:class:`Proxy` class is a singleton dictionary storing
    proxy settings.
    It typically maps each protocol (e.g., "http", "https") with
    the corresponding proxy URL (e.g., "http://localhost:8080").

    >>> proxy = Proxy()
    >>> proxy_enable("localhost", 8080)
    >>> proxy_disable()
    """
    pass

def proxy_enable(host :str, port :int, protocols :list = None):
    """
    Enables the :py:class:`Proxy` singleton.

    Args:
        host (str): The proxy FQDN or IP address.
        port (int): The proxy port.
        protocols (list): The list of protocols supported by the proxy.
            Passing ``None`` is equivalent to passing ``["http", "https"]``.
            Defaults to ``None``.
    """
    if not protocols:
        protocols = ["http", "https"]
    proxy = Proxy()
    for protocol in protocols:
        url = f"{protocol}://{host}:{port}"
        proxy[protocol] = url

def proxy_disable():
    """
    Disables the :py:class:`Proxy` singleton.
    """
    proxy = Proxy()
    proxy.clear()

def make_session() -> requests.Session:
    """
    Creates a :py:class:`requests.Session` instance according to
    the :py:class:`Proxy` singleton.

    Returns:
        The corresponding :py:class:`requests.Session` instance.
    """
    session = requests.Session()
    proxy = Proxy()
    if proxy:
        Log.info("Setting proxy = %s" % proxy)
        session.proxies.update(proxy)
    return session

#---------------------------------------------------------------
# Example to use a local proxy
#---------------------------------------------------------------

PROXY_LOCALHOST_IP    = "127.0.0.1"
PROXY_LOCALHOST_PORT  = 8080

def proxy_enable_localhost():
    """
    Illustrative function enabling a local proxy (127.0.0.1:8080).
    """
    proxy_enable(PROXY_LOCALHOST_IP, PROXY_LOCALHOST_PORT)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

from .connector import Connector
from .query import Query, ACTION_READ, action_to_str

class EntriesConnector(Connector):
    """
    :py:class:`EntriesConnector` wraps a list of minifold entries
    (list of dictionaries)
    """
    def __init__(self, entries: list):
        """
        Constructor.

        Args:
            entries (list): A list of minifold entries.
        """
        super().__init__()
        self.m_keys = set()
        if len(entries) > 0:
            for entry in entries:
                self.m_keys |= set(entry.keys())
        self.m_entries = entries

    def attributes(self, obj: str = None) -> set:
        """
        Lists available attributes related to a given collection of object
        stored in this :py:class:`EntriesConnector` instance.

        Args:
            obj (str): The name of the collection of entries.
                As :py:class:`EntriesConnector` manages a single connection
                you may pass ``None``.

        Returns:
            The set of available attributes.
        """
        return set(self.m_keys)

    def query(self, query: Query) -> list:
        """
        Handles an input :py:class:`Query` instance.

        Args:
            query (Query): The handled query.

        Returns:
            The list of entries matching the input Query.
        """
        super().query(query)
        if query.action == ACTION_READ:
            ret = self.reshape_entries(query, self.entries)
        else:
            action = action_to_str(query.action)
            raise RuntimeError(
                f"EntriesConnector.query: {action} not yet implemented"
            )
        return self.answer(query, ret)

    @property
    def entries(self) -> list:
        """
        Accessor to the entries nested in this :py:class:`EntriesConnector` instance.

        Returns:
            The nested entries.
        """
        return self.m_entries



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

import os, sys
from .cache import DEFAULT_CACHE_STORAGE_BASE_DIR, DEFAULT_CACHE_STORAGE_LIFETIME
from .filesystem import check_writable_directory, mkdir

try:
    import requests_cache
except ImportError as e:
    from .log import Log
    Log.warning(
        "Please install requests-cache.\n"
        "  APT: sudo apt install python3-requests-cache\n"
        "  PIP: sudo pip3 install --upgrade requests-cache\n"
    )
    raise e

def install_cache(cache_filename :str = None):
    """
    Enables ``requests_cache`` for minifold, hence allowing to cache HTTP queries
    issued by minifold.

    Args:
        cache_filename (str): The path to the minifold cache. You may pass
            ``None`` to use the default path (i.e., ``~/.minifold/cache/requests_cache``
            under Linux).
    """
    if not cache_filename:
        directory = DEFAULT_CACHE_STORAGE_BASE_DIR
        cache_filename = os.path.join(DEFAULT_CACHE_STORAGE_BASE_DIR, "requests_cache")
    else:
        directory = os.path.dirname(cache_filename)
    mkdir(directory)
    check_writable_directory(directory)
    requests_cache.install_cache(
        cache_filename,
        expire_after = DEFAULT_CACHE_STORAGE_LIFETIME
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

from enum import IntEnum

class DocType(IntEnum):
    """
    The :py:class:`DocType` class enumerates the different kind of
    scientific publication.

    See also:

    - :py:class:`HalConnector`
    - :py:class:`DblpConnector`
    """
    BOOKS_AND_THESES = 0     # Books and Theses
    ARTICLE          = 1     # ART, Conference and Workshop Papers
    COMMUNICATION    = 2     # COMM
    JOURNAL          = 3     # Journal Articles
    REPORT           = 4     # Report, Informal Publications
    POSTER           = 5     # POSTER
    HDR              = 6     # HDR
    CHAPTER          = 7     # COUV, Editorship
    PATENT           = 8     # PATENT
    UNKNOWN          = 9999  # UNDEFINED or ""

    def __str__(self) -> str:
        """
        Converts a :py:class:`DocType` instance to the corresponding string.

        Returns:
            The corresponding string representation.
        """
        s = IntEnum.__str__(self)
        class_name = "%s" % self.__class__.__name__
        s = s[len(class_name)+1:]
        s = s.replace("_", " ")
        return s.title()

    def __lt__(self, other) -> bool:
        """
        Compares this :py:class:`DocType` instance to another one.

        Args:
            other (DocType): Another :py:class:`DocType` instance.

        Returns:
            ``True`` if ``self`` is lower than ``other``.
        """
        return self.value < other.value

def doc_type_to_html(t: DocType, n: int = 1) -> str:
    """
    Builds the HTML label related to a collection of homogeneous documents.

    Args:
        t (DocType): The type of the documents.
        n (int): The number of document.

    Returns:
        The corresponding HTML label.
    """
    ret = (
        "Misc"            if t == DocType.UNKNOWN else
        "Editorship"      if t == DocType.CHAPTER else
        "Journal article" if t == DocType.JOURNAL else
        t
    )

    # Add "s" at the end if needed
    if n > 1 and t not in {
        DocType.BOOKS_AND_THESES,
        DocType.HDR,
        DocType.CHAPTER,
        DocType.UNKNOWN
    }:
        ret = f"{ret}s"

    return ret



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

"""
This file provides utilities to wrap the ``pycountry`` python module.
"""

from .log import Log

try:
    import pycountry
except ImportError as e:
    from .log import Log
    Log.warning(
        "\n".join([
            "Please install requests",
            "  APT: sudo apt install python3-pycountry",
            "  PIP: sudo pip3 install --upgrade pycountry",
        ])
    )
    raise e

def _country_code_to_name(country_code :str) -> str:
    """
    Fixes an obsolete country code.

    Args:
        country_code (str): The input country code.

    Returns:
        The fixed country code.
    """
    # For obsolete version of pycountry
    Log.warning("Please update python3-pycountry; apt-get update && apt-get upgrade")

    ret = None

    try:
        country = pycountry.countries.get(alpha_2 = country_code.upper())
        ret = country.name
    except KeyError as e:
        Log.warning("Unknown country %r" % country_code)

    return ret

def country_code_to_name(country_code :str) -> str:
    """
    Retrieves the name of a country given its code.

    Args:
        country_code (str): The input country code.

    Returns:
        The corresponding country name.
    """
    if country_code is None:
        return None

    ret = None
    try:
        country = pycountry.countries.get(alpha_2 = country_code.upper())
        if country is None:
            return None
        ret = country.name
    except KeyError as e:
        if "%s" % e == "alpha_2":
            ret = _country_code_to_name(country_code)
        else:
            Log.warning("Unknown country %r" % country_code)

    return ret



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

import re
import string
import unicodedata

MAP_TO_INTERNATIONAL = {
    # Upper case
    "Á": "A",
    "À": "A",
    "Ä": "A",
    "Ć": "C",
    "Ç": "C",
    "Đ": "D",
    "É": "E",
    "Ê": "E",
    "È": "E",
    "Ë": "E",
    "Í": "I",
    "Î": "I",
    "Ï": "I",
    "Ł": "L",
    "Ö": "O",
    "Ô": "O",
    "Û": "U",
    "Ü": "U",
    "Š": "S",
    # Lower case
    "á": "a",
    "à": "a",
    "ä": "a",
    "ć": "c",
    "ç": "c",
    "đ": "d",
    "é": "e",
    "ê": "e",
    "è": "e",
    "ë": "e",
    "í": "i",
    "î": "i",
    "ï": "i",
    "ł": "l",
    "ö": "o",
    "ô": "o",
    "û": "u",
    "ü": "u",
    "š": "s",
    # Only lower
    "ß": "ss",
}

def to_international_chr(c: str) -> str:
    """
    Converts international characters to the corresponding character(s)
    in ``a-zA-Z``.

    Examples:
        >>> to_international_chr("ß")
        'ss'
        >>> to_international_chr("é")
        'e'

    Args:
        c (str): The input character.

    Returns:
        The corresponding characters(s).
    """
    return MAP_TO_INTERNATIONAL.get(c, c)

def to_international_string(s: str) -> str:
    """
    Converts a string involving international characters
    (see `:func:py:to_international_chr`) to the corresponding
    string.

    Example:
        >>> to_international_string("élève")
        'eleve'

    Args:
        s (str): The input string.

    Returns:
        The converted string.
    """
    translator = str.maketrans(MAP_TO_INTERNATIONAL)
    return s.translate(translator)

def remove_punctuation(s: str) -> str:
    """
    Replaces the punctuation characters by spaces.

    Example:
        >>> remove_punctuation("Example: a sentence, with punctuation."
        'Example  a sentence  with punctuation '

    Args:
        s (str): The input string.

    Returns:
        The converted string.
    """
    translator = str.maketrans({key: " " for key in string.punctuation})
    return s.translate(translator)

def remove_html_tags(s: str) -> str:
    """
    Removes the HTML tag from an HTML string.

    Example:
        >>> remove_html_tags("<a href='#'>A link</a> <u><i>italic underlined</i><u>")
        'A link bold italic underlined'

    Args:
        s (str): The input string.

    Returns:
        The converted string.
    """
    return re.sub("<[^>]+>", "", s)

def remove_html_escape_sequences(s: str) -> str:
    """
    Removes the HTML sequences.

    Example:
        >>> remove_html_escape_sequences("cha&icirc;ne")
        'chane'

    Args:
        s (str): The input string.

    Returns:
        The converted string.
    """
    return re.sub("[&].*;", "", s)

def remove_latex_escape_sequence(s: str) -> str:
    """
    Removes the latex sequences.

    Example:
        >>> remove_latex_escape_sequence("n\\oe ud")
        'nud'

    Args:
        s (str): The input string.

    Returns:
        The converted string.
    """
    return re.sub("\\\\[a-zA-Z0-9]+ ", "", s)

def to_canonic_string(s: str) -> str:
    """
    Canonizes a string.

    Args:
        s (str): The input string.

    Returns:
        The converted string.
    """
    s = s.lower()
    s = to_international_string(s)
    s = remove_punctuation(s)  # This includes '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~'
    s = remove_latex_escape_sequence(s)
    s = s.replace(" ", "")  # Remove  spaces
    return s

def to_canonic_fullname(s: str) -> str:
    """
    Canonizes a fullname.

    Args:
        s (str): The input fullname.

    Returns:
        The converted fullname.
    """
    s = to_international_string(s.lower())
    s = s.replace("-", " ")
    s = re.sub("[(].+[)]", "", s)      # Remove surnames
    s = remove_latex_escape_sequence(s)
    s = remove_html_tags(s)
    s = remove_html_escape_sequences(s)
    s = re.sub("[a-zA-Z]+[.]", "", s)  # Remove sigles
    s = " ".join(s.split())            # Remove useless spaces
    return s

def unicode_to_utf8(s: str) -> str:
    """
    Converts an unicode string to an UTF-8 string.

    Args:
        s (str): The input unicode string.

    Returns:
        The corresponding UTF-8 string.
    """
    return unicodedata.normalize("NFKD", s)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

"""
Internals to make hashable object that are not hashable.

This is useful, e.g, in the :py:class:`UniqueConnector` class.
"""

def to_hashable(x):
    """
    Converts an object to an hashable ``tuple``.

    Args:
        x (object): The input object.

    Returns:
        The corresponding hashable object.
    """
    return tuple(to_hashable(elt) for elt in x) if isinstance(x, (set, list)) \
      else tuple((to_hashable(k), to_hashable(v)) for k, v in x.items()) if isinstance(x, dict) \
      else x



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

# ---------------------------------------------------------------------------------
# Requests
# ---------------------------------------------------------------------------------

import sys

if sys.version_info < (3, 5):  # async, await
    raise RuntimeError("asyncio requires python>=3.5")

try:
    import requests
except ImportError as e:
    from .log import Log
    Log.warning(
        "\n".join([
            "Please install requests",
            "  APT: sudo apt install python3-requests",
            "  PIP: sudo pip3 install --upgrade requests",
        ])
    )
    raise e

import asyncio, datetime, re
import concurrent.futures

from copy import deepcopy
from functools import partial
from .connector import Connector
from .query import ACTION_READ, Query
from .html import html_to_text
from .log import Log
from .proxy import make_session
from .request_cache import install_cache

DEFAULT_TIMEOUT = (1.0, 2.0)    # (connect timeout, read timeout)
# DEFAULT_TIMEOUT = (0.5, 1.0)  # (connect timeout, read timeout)

def now() -> str:
    """
    Crafts a string storing the current timestamp.

    Returns:
        A string containing current timestamp.
    """
    return datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

def download(url: str, timeout: tuple = DEFAULT_TIMEOUT, cache_filename: str = None):
    """
    Downloads the content related to a given URL.
    See also :py:func:`request_cache` to enable caching.

    Args:
        url (str): A string containing the target URL.
        timeout (tuple): A ``(float, float)`` tuple corresponding to the
            (connect timeout, read timeout).
        cache_filename (str): A string containing path to the cache to use.
            Pass ``None`` to use the default cache.

    Raises:
        requests.exceptions.ConnectionError
        requests.exceptions.ConnectTimeout
        requests.exceptions.ContentDecodingError
        requests.exceptions.ReadTimeout
        requests.exceptions.SSLError

    Returns:
        The corresponding response.
    """
    install_cache(cache_filename)
    try:
        # Log.debug("download: GET %s (timeout = %s)" % (url, timeout))
        # https://stackoverflow.com/questions/8287628/proxies-with-python-requests-module
        session = make_session()
        return session.get(
            url,
            timeout = timeout
        )
    except Exception as exc:
        Log.warning("download: GET %s (timeout = %s): %s" % (url, timeout, exc))
        return exc

def trim_http(url: str) -> str:
    """
    Discards the ``"http://"`` or `` "https://"`` prefix from an url.

    Args:
        url (str): A str containing an URL.

    Returns:
        The str corresponding to the trimmed URL.
    """
    return re.sub("https?://", "", url, flags = re.IGNORECASE)

# https://skipperkongen.dk/2016/09/09/easy-parallel-http-requests-with-python-and-asyncio/
async def downloads_async(
    urls: list,
    timeout: tuple = DEFAULT_TIMEOUT,
    return_exceptions: bool = True
):
    """
    Asynchronous download procedure.

    Args:
        urls (list): An iterable over strings, each of them corresponding to an URL.
        timeout (tuple): A tuple ``(float, float)`` corresponding to the
            (connect timeout, read timeout).
        return_exceptions (bool): Pass ``True`` if this function is allowed to raise
            exceptions or must be quiet, ``False`` otherwise. Defaults to ``True``.

    Raises:
        See the :py:func:`download` function for the list of possible exceptions.

    Returns:
        A ``dict({str : ?}}`` mapping for each queried URL the corresponding
        contents (if successful), the corresponding :py:class:`Exception` otherwise.
    """
    with concurrent.futures.ThreadPoolExecutor(max_workers=500) as executor:
        loop = asyncio.get_event_loop()
        futures = [
            loop.run_in_executor(
                executor,
                partial(download, timeout=timeout, cache_filename=None),
                url
            ) for url in urls
        ]
        return {
            url : response
            for (url, response) in zip(
                urls,
                await asyncio.gather(*futures, return_exceptions = return_exceptions)
            )
        }

def downloads(*args) -> dict:
    """
    Performs multiple downloads in parallel. See downloads_async for further details.
    See also :py:func:`request_cache` to enable caching.

    Args:
        urls (list): An iterable over strings, each of them corresponding to an URL.
        timeout (tuple): A tuple ``(float, float)`` corresponding to the
            (connect timeout, read timeout).
        return_exceptions (bool): Pass ``True`` if this function is allowed to raise
            exceptions or must be quiet, ``False`` otherwise. Defaults to ``True``.

    Returns:
        A ``dict({str : ?}}`` mapping for each queried URL the corresponding
        contents (if successful), the corresponding :py:class:`Exception` otherwise.
    """
    Log.debug("[%s] downloads: start" % now())
    # https://stackoverflow.com/questions/46727787/runtimeerror-there-is-no-current-event-loop-in-thread-in-async-apscheduler
    # loop = asyncio.get_event_loop()
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    ret = loop.run_until_complete(downloads_async(*args))
    Log.debug("[%s] downloads: %d URLs fetched" % (now(), len(ret)))
    return ret

# ---------------------------------------------------------------------------------
# Minifold
# ---------------------------------------------------------------------------------

def extract_response(response: object, extract_text: bool = True) -> object:
    """
    Extracts from a response the corresponding contents or Exception.

    Args:
        response (object): A :py:class:`requests.Response` or an
            :py:class:`Exception` instance.
        extract_text (bool): A bool indicating if text must be extracted
            from HTML.

    Returns:
        The corresponding Exception or str.
    """
    if isinstance(response, Exception):
        return response  # Forward exception
    text = response.text
    return html_to_text(text) if extract_text else text

RE_VALID_URL = re.compile("https?://.*")

class DownloadConnector(Connector):
    """
    The :py:class:`DownloadConnector` allows to fetch data over HTTP.
    """
    def __init__(
        self,
        map_url_out: dict,
        child: Connector,
        downloads :callable = downloads,
        extract_response: callable = extract_response
        # TODO Pass downloads() parameters
    ):
        """
        Constructor.

        Args:
            map_url_out (dict): dict(str : str) entry attribute containing an URL to another
                entry attribute that will store the contents provided by this URL.
                Example: ``{"url" : "html_content"}``
            child (Connector): The child Connector.
            downloads (callable): ``Callback(urls) -> dict(url : content)`` where
                urls is an iterable of URLs and where.
                the returned dict maps each urls and the corresponding response.
                Note: You could pass ``partial(download, ...)`` to customize the timeouts.
            extract_response (callable): ``Callback(response) -> str`` callback used to extract
                data from an HTTP query.
        """
        super().__init__()
        self.map_url_out = map_url_out
        self.child = child
        self.downloads = downloads
        self.extract_response = extract_response

    def query(self, query :Query) -> list:
        """
        Handles an input :py:class:`Query` instance..

        Args:
            query: A :py:class:`Query` instance.

        Returns:
            The corresponding results.
        """
        if query.action == ACTION_READ:
            # Pull child records
            q_child = deepcopy(query)  # TODO: rework q_child like in LambdasConnector
            q_child.attributes = set(query.attributes) - set(self.map_url_out.values())
            entries = self.child.query(q_child)

            attributes = set(query.attributes) & set(self.map_url_out.values()) \
                if query.attributes else set(self.map_url_out.values())

            # If some URLs must be queried
            if attributes:
                # Fetch URLs
                urls = set()
                for attr_url in self.map_url_out.keys():
                    urls |= {
                        entry.get(attr_url)
                        for entry in entries
                        if (
                            isinstance(entry.get(attr_url), str)
                            and RE_VALID_URL.match(entry.get(attr_url))
                        )
                    }
                Log.debug("Starting fetching %d URLs" % len(urls))
                map_url_response = self.downloads(urls)
                if self.extract_response:
                    map_url_response = {
                        url : self.extract_response(response)
                        for (url, response) in map_url_response.items()
                    }

                # Dispatch responses in appropriate entry attributes
                for (attr_url, attr_out) in self.map_url_out.items():
                    for entry in entries:
                        url = entry.get(attr_url)
                        if url:
                            entry[attr_out] = map_url_response[url]

            # Forward to parent Connector once entry are reshaped.
            entries = self.reshape_entries(query, entries)
            return entries
        else:
            raise RuntimeError("Action not implemented: %s" % query.action)

    def attributes(self, object: str) -> set:
        """
        Lists available attributes related to a given collection of object
        stored in this :py:class:`DownloadConnector` instance.

        Args:
            object (str): The name of the collection of entries.

        Returns:
            The set of available attributes for ``object``.
        """
        return set(self.child.attributes(object)) | set(self.map_url_out.values())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

def reverse_dict(d: dict) -> dict:
    """
    Reverses a dictionary (swap its key and its values).
    Note that (key, value) pair may disappear if the values
    are not unique.

    Args:
        d (dict): The input dictionary.

    Returns:
        The reversed dictionary.
    """
    return {v : k for (k, v) in d.items()}

def freeze_dict(d: dict) -> dict:
    """
    Freezes the values of a dictionary (for each value of type ``set``).

    Args:
        d (dict): The input dictionary.

    Returns:
        The freezed dictionary.
    """
    return {k : frozenset(v) if isinstance(v, set) else v for (k, v) in d.items()}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

"""
This file gathers some useful function to get and display
some entries from an abitrary :py:class:`Connector` instance.
"""

from .connector import Connector
from .html import html
from .query import Query
from .unique import UniqueConnector

def get_values(connector: Connector, attribute: str) -> list:
    """
    Retrieves distinct values mapped with a given attribute stored
    in the entries of a given :py:class:`Connector` instance.

    Args:
        connector (Connector): The queried :py:class:`Connector` instance.
        attribute (str): The queried attribute.

    Raises:
        KeyError: if ``attribute`` is not a valid attribute.

    Returns:
        The corresponding values.
    """
    enriched_connector = UniqueConnector(
        [attribute],
        connector
    )
    q = Query(attributes = [attribute])
    return [entry[attribute] for entry in enriched_connector.query(q)]

def show_some_values(
    connector :Connector,
    limit :int = 5,
    max_strlen :int = None,
    query :Query = None
):
    """
    Displays in Jupyter notebook some values for each attribute of the entries
    served by a given :py:class:`Connector` instance.
    This is useful to discover a dataset.

    Args:
        connector (Connector): The queried :py:class:`Connector` instance.
        limit (int): A positive integer or ``None`` (no limit) that upper bounds
            the number of values to be fetched for each attribute.
        query (Query): The :py:class:`Query` instance used to probe ``connector``.
            You may pass ``None`` to issue to default query.
    """
    if not query:
        query = Query()

    map_values = {
        attribute : get_values(connector, attribute)
        for attribute in connector.attributes(query.object)
    }

    html(
        """
        <table>
            <tr>
                <th>Attribute</th>
                <th>#values</th>
                <th>%(limit)d first values</th>
            </tr>
            %(rows)s
        </table>
        """ % {
            "limit" : limit,
            "rows"  : "".join(["""
                <tr>
                    <td><code>%(attribute)s</code></td>
                    <td>%(num_values)d</td>
                    <td style="text-align:left"><code>%(values)s</code></td>
                </tr>
                """ % {
                    "attribute"  : attribute,
                    "num_values" : len(map_values[attribute]),
                    "values" : "<br/>".join([
                        str(x)[:max_strlen] for x in map_values[attribute][:limit]
                    ]),
                } for attribute in connector.attributes(query.object)
            ]),
        }
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

from itertools import chain
from .connector import Connector
from .query import Query, ACTION_READ

def union_gen(list_entries: list) -> iter:
    """
    Iterator over a list of list of dicts.

    Args:
        list_entries: A list of lists of dicts.

    Returns:
        The resulting generator.
    """
    return chain(*list_entries)

def union(list_entries: list) -> list:
    """
    Make the union of several lists of entries.
    Use minifold.distinct() to remove duplicates.

    Args:
        list_entries: A list of lists of dicts.

    Returns:
        The resulting list of dicts.
    """
    return [entry for entry in union_gen(list_entries)]

class UnionConnector(Connector):
    """
    The :py:class:`UnionConnector` class implements the UNION
    statement in a minifold pipeline.
    """
    def __init__(self, children: list):
        """
        Constructor.

        Args:
            child (Connector): The list of children
                minifold :py:class:`Connector` instances.
        """

        super().__init__()
        self.children = children

    def attributes(self, object: str) -> set:
        """
        Lists the available attributes related to a given collection of
        minifold entries exposed by this :py:class:`WhereConnector` instance.

        Args:
            object (str): The name of the collection.

        Returns:
            The set of corresponding attributes.
        """
        attrs = set()
        for child in self.children:
            try:
                attrs |= child.attributes(object)
            except:
                pass
        return attrs

    def query(self, query :Query) -> list:
        """
        Handles an input :py:class:`Query` instance.

        Args:
            query (Query): The handled query.

        Returns:
            The list of entries matching the input query.
        """
        super().query(query)
        if query.action != ACTION_READ:
            raise ValueError("Invalid action in query %s" % query)
        return union([child.query(query) for child in self.children])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

"""
This file gathers some utilities to process, render, or make HTML
strings in minifold.
"""

import re, sys
from html import escape
from .connector import Connector
from .query import Query

# Regular expression matching an URL
PATTERN_URL = re.compile(
    "http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*(),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+"
)

#def to_html_link(d: dict, k: str, v: str) -> str:
#    try:
#        return "<a href=\"%s\">%s</a>" % (d[k], escape(v))
#    except KeyError:
#        return v

def value_to_html(x: object) -> str:
    """
    Converts a value to its corresponding HTML string.

    Args:
        x (object): The input value.

    Returns:
        The corresponding HTML string.
    """
    def str_to_html(s :str) -> str:
        """
        Right-strips ``"."`` and ``","`` from a string.

        Args:
            s (str): The input string.
        """
        return s.rstrip(".").rstrip(",")

    if isinstance(x, dict):
        ret = dict_to_html(x, x.keys())
    elif isinstance(x, str):
        if PATTERN_URL.match(x):
            ret = "<a href=\"%s\">%s</a>" % (x, escape(x))
        else:
            ret = str_to_html(x)
    elif isinstance(x, (list, set)):
        ret = ", ".join([value_to_html(value) for value in x])
    else:
        ret = escape("%s" % x)
    return ret

def html(s :str):
    """
    Evaluates HTML code in a Jupyter Notebook.

    Args:
        s: A str containing HTML code.
    """
    from IPython.display import display, HTML
    chart = HTML(s)
    # or chart = charts.plot(...)
    display(chart)

def print_error(x: object):
    """
    Prints an error in a Jupyter Notebook.

    Args:
        x: An expection.
    """
    print(str(x), file = sys.stderr)

def entries_to_html(
    entries :list,
    map_attribute_label :dict = None,
    attributes :list = None,
    keep_entry_if :callable = None
) -> str:
    """
    Exports to HTML a list of dict.

    Args:
        entries: A list of dicts
        map_attribute_label: A dict {str : str} which maps each entry key with the column header
            to display.
        attributes: The subset of keys to display.
        keep_entry_if: Callback allowing to filter some entries

    Returns:
        The corresponding HTML string.
    """
    if not attributes:
        attributes = list(entries[0].keys())
    if keep_entry_if is None:
        def keep_entry_if(i, entry):
            return True
    if not map_attribute_label:
        map_attribute_label = dict()
    return """
        <div style="width:100%%; overflow-x: auto;white-space: nowrap;">
            <table>
                %(header)s
                %(rows)s
            </table>
        </div>
        """ % {
            "header" : "<tr>%s</tr>" % "".join(
                ["<th>Index</th>"] + [
                    "<th style='text-align: left;white-space: nowrap;'>%s</th>"
                    % str(map_attribute_label.get(attribute, attribute))
                    for attribute in attributes
                ]
            ),
            "rows" : "".join([
                """
                <tr>
                    <td>%(index)d</td>
                    %(values)s
                </tr>
                """ % {
                    "index" : i,
                    "values" : "".join([
                        "<td style='text-align: left;white-space: nowrap;'>%s</td>"
                        % value_to_html(entry.get(attribute))
                        for attribute in attributes
                    ])
                }
                for (i, entry) in enumerate(entries)
                if keep_entry_if(i, entries)
            ]),
        }

def connector_to_html(connector :Connector, **kwargs) -> str:
    attributes = kwargs.get("attributes")
    if not attributes:
        attributes = connector.attributes(None)
    entries = connector.query(Query(**kwargs))
    return entries_to_html(entries, attributes=attributes)

def entry_to_html(entry :dict, map_attribute_label :dict = None, attributes :list = None) -> str:
    return entries_to_html([entry], map_attribute_label, attributes)

# OBSOLETE
def dict_to_html(d :dict, attributes :list, map_attribute_label :dict = None) -> str:
    return entry_to_html(d, map_attribute_label, attributes)

# ----------------------------------------------------------------------------------
# Extensions depending on BeautifulSoup
# ----------------------------------------------------------------------------------

try:
    # See https://matix.io/extract-text-from-webpage-using-beautifulsoup-and-python/
    from bs4 import BeautifulSoup

    HTML_BLACKLIST_TAGS = {
        "[document]",
        "noscript",
        "header",
        "html",
        "head",
        "link",
        "input",
        "meta",
        "script",
        "style"
    }

    def html_to_text(s_html: str, blacklist: set = None) -> str:
        """
        Convert an HTML page to text, by discarding javascript and css related
        to the site.

        Args:
            s_html (str): A str containing HTML.
            blacklist (set): A set of string (lowercase) corresponding to HTML tags
                that must be ignored.

        Returns:
            The corresponding text.
        """
        if blacklist is None:
            blacklist = HTML_BLACKLIST_TAGS
        soup = BeautifulSoup(s_html, features = "lxml")
        soup.find_all(s_html, "html.parser")
        text = soup.find_all(string = True)
        l = list()
        for t in text:
            s = t.strip()
            if s and t.parent.name not in blacklist:
                # print("<%s>: %r" % (t.parent.name, s))
                s = re.sub("<[^>]*>", "", s)
                l += [s]
        s = " ".join(l)
        return re.sub(r"[\n]+", "\n\n", s)

    HTML_MAP_WHITELIST_ATTRS = {
        "a"   : ["href"],
        "img" : ["src"],
    }

    # Based on https://gist.github.com/revotu/21d52bd20a073546983985ba3bf55deb
    def remove_all_attrs_except_saving(soup, whitelist = None):
        """
        Remove all attributes except some.

        Args:
            soup: A BeautifulSoup instance, modified in place.
            whitelist: A dict {tag : list(attr)} where tag is an HTML tag and attr
                an HTML attribute.
        """
        if whitelist is None:
            whitelist = HTML_MAP_WHITELIST_ATTRS
        for tag in soup.find_all(True):
            if tag.name not in whitelist.keys():
                tag.attrs = {}
            else:
                attrs = dict(tag.attrs)
                allowed_attrs = whitelist[tag.name]
                for attr in attrs:
                    if attr not in allowed_attrs:
                        del tag.attrs[attr]
                    elif attr == "href" and not attr.startswith("http"):
                        tag.attrs[attr] = "#"
        return soup

    def remove_tags(soup, blacklist :set = None):
        """
        Remove some HTML tags.

        Args:
            soup: A BeautifulSoup instance, modified in place.
            blacklist: A list of str, where each str is an HTML tag.
        """
        if blacklist is None:
            blacklist = HTML_BLACKLIST_TAGS
        for tag in blacklist:
            while True:
                attr = getattr(soup, tag)
                if attr:
                    attr.decompose()
                else:
                    break

    def sanitize_html(
        s_html       :str,
        blacklist    :set  = None,
        remove_attrs :bool = True
    ) -> str:
        """
        Remove from an HTML string irrelevant HTML blocks and attributes.

        !!! This function is SLOW so do not use it on large corpus!

        Args:
            s_html: A str instance containing HTML.
            blacklist: List of blacklisted HTML tags.
            remove_attrs:

        Returns:
            The sanitized string.
        """
        if blacklist is None:
            blacklist = HTML_BLACKLIST_TAGS
        soup = BeautifulSoup(s_html, features = "lxml")
        if remove_attrs:
            remove_all_attrs_except_saving(soup)
        body = soup.find("body")
        if not body:
            return None
        remove_tags(body, blacklist)
        return str(body.prettify())

except ImportError as e:
    from .log import Log
    Log.warning(
        "Please install pymongo.\n"
        "  APT: sudo apt install python3-bs4\n"
        "  PIP: sudo pip3 install --upgrade bs4\n"
    )
    raise e

